# #!/usr/bin/env python3
# """
# # NAME:     RMpeakfit_3D.py                                                   #
# #                                                                             #
# # PURPOSE:  Fit peak of RM spectra, for every pixel in 3D FDF cube.           #
# #
# # Initial version: Cameron Van Eck, Dec 2020
# """

# import os
# import sys

# import astropy.io.fits as pf
# import numpy as np
# from tqdm.auto import trange

# from rm_lite.tools_3d.rmsynth import readFitsCube, readFreqFile
# from rm_lite.utils.fitting import interp_images
# from rm_lite.utils.synthesis import fits_make_lin_axis, measure_FDF_parms

# C = 2.997924538e8  # Speed of light [m/s]


# def pixelwise_peak_fitting(
#     FDF,
#     phiArr,
#     fwhmRMSF,
#     lamSqArr_m2,
#     lam0Sq,
#     product_list,
#     noiseArr=None,
#     stokesIcube=None,
#     weightType="uniform",
# ):
#     """
#     Performs the 1D FDF peak fitting used in RMsynth/RMclean_1D, pixelwise on
#     all pixels in a 3D FDF cube.

#     Inputs:
#         FDF: FDF cube (3D array). This is assumed to be in astropy axis ordering
#             (Phi, dec, ra)
#         phiArr: (1D) array of phi values
#         fwhmRMSF: 2D array of RMSF FWHM values
#         lamSqArr_m2: 1D array of channel lambda^2 values.
#         lam0Sq: scalar value for lambda^2_0, the reference wavelength squared.
#         product_list: list containing the names of the fitting products to save.
#         dFDF: 2D array of theoretical noise values. If not supplied, the
#             peak fitting will default to using the measured noise.
#     Outputs: dictionary of 2D maps, 1 per fit output
#     """

#     # FDF: output by synth3d or clean3d
#     # phiArr: can be generated from FDF cube
#     # fwhm: 2D map produced synth3D
#     # dFDFth: not currently produced (default mode not to input noise!)
#     #   If not present, measure_FDF_parms uses the corMAD noise.
#     #
#     # lamSqArr is only needed for computing errors in derotated angles
#     #   This could be compressed to a map or single value from RMsynth?
#     # lam0Sq is necessary for de-rotation

#     map_size = FDF.shape[1:]

#     # Create pixel location arrays:
#     xarr, yarr = np.meshgrid(range(map_size[0]), range(map_size[1]))
#     xarr = xarr.ravel()
#     yarr = yarr.ravel()

#     # Create empty maps:
#     map_dict = {}
#     for parameter in product_list:
#         map_dict[parameter] = np.zeros(map_size)

#     freqArr_Hz = C / np.sqrt(lamSqArr_m2)
#     freq0_Hz = C / np.sqrt(lam0Sq)
#     if stokesIcube is not None:
#         idx = np.abs(freqArr_Hz - freq0_Hz).argmin()
#         if freqArr_Hz[idx] < freq0_Hz:
#             Ifreq0Arr = interp_images(
#                 stokesIcube[idx, :, :], stokesIcube[idx + 1, :, :], f=0.5
#             )
#         elif freqArr_Hz[idx] > freq0_Hz:
#             Ifreq0Arr = interp_images(
#                 stokesIcube[idx - 1, :, :], stokesIcube[idx, :, :], f=0.5
#             )
#         else:
#             Ifreq0Arr = stokesIcube[idx, :, :]
#     else:
#         Ifreq0Arr = np.ones(map_size)
#         stokesIcube = np.ones((freqArr_Hz.size, map_size[0], map_size[1]))

#     # compute weights if needed:
#     if noiseArr is not None:
#         if weightType == "variance":
#             weightArr = 1.0 / np.power(noiseArr, 2.0)
#             weightArr = np.where(np.isnan(weightArr), 0.0, weightArr)
#         elif weightType == "uniform":
#             weightArr = np.ones(lamSqArr_m2.shape, dtype=np.float32)
#             weightArr = np.where(np.isnan(noiseArr), 0.0, weightArr)
#         else:
#             raise Exception("Invalid weight type; must be 'uniform' or 'variance'")

#         dFDF = Ifreq0Arr * np.sqrt(
#             np.sum(weightArr**2 * np.nan_to_num(noiseArr) ** 2)
#             / (np.sum(weightArr)) ** 2
#         )

#     else:
#         weightArr = np.ones(lamSqArr_m2.shape, dtype=np.float32)
#         dFDF = None

#     # Run fitting pixel-wise:
#     for i in trange(xarr.size):
#         FDF_pix = FDF[:, xarr[i], yarr[i]]
#         fwhmRMSF_pix = fwhmRMSF[xarr[i], yarr[i]]
#         if type(dFDF) == type(None):
#             dFDF_pix = None
#         else:
#             dFDF_pix = dFDF[xarr[i], yarr[i]]
#         try:
#             mDict = measure_FDF_parms(
#                 FDF_pix,
#                 phiArr,
#                 fwhmRMSF_pix,
#                 dFDF=dFDF_pix,
#                 lamSqArr_m2=lamSqArr_m2,
#                 lam0Sq=lam0Sq,
#                 snrDoBiasCorrect=5.0,
#             )
#             # Add keywords not included by the above function:
#             mDict["lam0Sq_m2"] = lam0Sq
#             mDict["freq0_Hz"] = freq0_Hz
#             mDict["fwhmRMSF"] = fwhmRMSF_pix
#             mDict["Ifreq0"] = Ifreq0Arr[xarr[i], yarr[i]]
#             mDict["fracPol"] = mDict["ampPeakPIfit"] / mDict["Ifreq0"]
#             mDict["min_freq"] = float(np.min(freqArr_Hz))
#             mDict["max_freq"] = float(np.max(freqArr_Hz))
#             mDict["N_channels"] = lamSqArr_m2.size
#             mDict["median_channel_width"] = float(np.median(np.diff(freqArr_Hz)))
#             if dFDF_pix is not None:
#                 mDict["dFDFth"] = dFDF_pix
#             else:
#                 mDict["dFDFth"] = np.nan
#             for parameter in product_list:
#                 map_dict[parameter][xarr[i], yarr[i]] = mDict[parameter]
#         except:
#             for parameter in product_list:
#                 map_dict[parameter][xarr[i], yarr[i]] = np.nan

#     return map_dict


# def delete_FITSheader_axis(fitsheader, axis_number):
#     """Deletes FITS keywords associated with the specified axis."""
#     axis_keywords = ["NAXIS", "CRVAL", "CRPIX", "CDELT", "CUNIT", "CTYPE"]
#     axis_str = str(axis_number)
#     for keyword in axis_keywords:
#         try:
#             del fitsheader[keyword + axis_str]
#         except:
#             pass


# def save_maps(map_dict, prefix_path, FDFheader):
#     """
#     Saves the selected 2D maps of the fit output.
#     Inputs:
#         map_dict: a dictionary of 2D maps for the fitting outputs.
#         prefix_path: the full or relative path to save the files, plus the file
#             prefix to be given to the files.
#     """
#     # Set up generic FITS header
#     product_header = FDFheader.copy()
#     product_header["NAXIS"] = 2
#     # Remove extra axes:
#     if "NAXIS3" in product_header:
#         delete_FITSheader_axis(product_header, 3)
#     if "NAXIS4" in product_header:
#         delete_FITSheader_axis(product_header, 4)
#     if "STOKES" in product_header:
#         del product_header["STOKES"]
#     product_header["HISTORY"] = (
#         "Polarization peak maps created with RM-Tools RMpeakfit_3D"
#     )

#     # Set flux unit from FITS header if possible
#     if "BUNIT" in FDFheader:
#         flux_unit = FDFheader["BUNIT"]
#     else:
#         flux_unit = ""

#     # Dictionary of units for peak fitting output parameters (for FITS headers)
#     unit_dict = {
#         "dFDFcorMAD": flux_unit,
#         "phiPeakPIfit_rm2": "rad/m^2",
#         "dPhiPeakPIfit_rm2": "rad/m^2",
#         "ampPeakPIfit": flux_unit,
#         "ampPeakPIfitEff": flux_unit,
#         "dAmpPeakPIfit": flux_unit,
#         "snrPIfit": "",
#         "indxPeakPIfit": "",
#         "peakFDFimagFit": flux_unit,
#         "peakFDFrealFit": flux_unit,
#         "polAngleFit_deg": "deg",
#         "dPolAngleFit_deg": "deg",
#         "polAngle0Fit_deg": "deg",
#         "dPolAngle0Fit_deg": "deg",
#         "Ifreq0": flux_unit,
#         "polyCoeffs": "",
#         "IfitStat": "",
#         "IfitChiSqRed": "",
#         "lam0Sq_m2": "m^2",
#         "freq0_Hz": "Hz",
#         "fwhmRMSF": "rad/m^2",
#         "dQU": flux_unit,
#         "dFDFth": flux_unit,
#         "min_freq": "Hz",
#         "max_freq": "Hz",
#         "N_channels": "",
#         "median_channel_width": "Hz",
#         "fracPol": "",
#         "sigmaAddQ": "",
#         "dSigmaAddMinusQ": "",
#         "dSigmaAddPlusQ": "",
#         "sigmaAddU": "",
#         "dSigmaAddMinusU": "",
#         "dSigmaAddPlusU": "",
#     }

#     # Check that directory exists (in case it is requested to save maps to new subdirectory):
#     if not os.path.exists(os.path.dirname(prefix_path)):
#         os.mkdir(os.path.dirname(prefix_path))

#     # per product, customize FITS header as needed and save file
#     for product in map_dict.keys():
#         if map_dict[product].dtype == np.float_:
#             data = map_dict[product].astype(np.float32)
#         else:
#             data = map_dict[product]
#         product_header["BUNIT"] = unit_dict[product]
#         pf.writeto(
#             prefix_path + product + ".fits", data, product_header, overwrite=True
#         )


# def find_axes(header):
#     """Identifies how many axes are present in a FITS file, and which is the
#     Faraday depth axis. Necessary for bookkeeping on cube dimensionality,
#     given that RM-clean only supports 3D cubes, but data may be 4D files."""
#     Ndim = header["NAXIS"]
#     FD_axis = Ndim
#     # Check for FD axes:
#     for i in range(1, Ndim + 1):
#         try:
#             if "FDEP" in header["CTYPE" + str(i)].upper():
#                 FD_axis = i
#         except:
#             pass  # The try statement is needed for if the FITS header does not
#             # have CTYPE keywords.
#     return Ndim, FD_axis


# def read_files(FDF_filename, freq_filename):
#     """
#     Read in the files needed for peak fitting. Files assumed to be the standard
#     outputs of RMsynth3D (and, optionally, RMclean3D). also, freq file.
#     """

#     HDUreal = pf.open(
#         FDF_filename.replace("_tot", "_real").replace("_im", "_real"),
#         "readonly",
#         memmap=True,
#     )
#     head = HDUreal[0].header.copy()
#     FDFreal = HDUreal[0].data

#     HDUimag = pf.open(
#         FDF_filename.replace("_tot", "_im").replace("_real", "_im"),
#         "readonly",
#         memmap=True,
#     )
#     FDFimag = HDUimag[0].data
#     complex_cube = FDFreal + 1j * FDFimag

#     # Identify Faraday depth axis (assumed to be last one if not explicitly found)
#     Ndim, FD_axis = find_axes(head)

#     # Move FD axis to first place in numpy order.
#     if FD_axis != Ndim:
#         complex_cube = np.moveaxis(complex_cube, Ndim - FD_axis, 0)

#     # Remove degenerate axes to prevent problems with later steps.
#     complex_cube = complex_cube.squeeze()

#     lam0Sq = head["LAMSQ0"]

#     freqArr_Hz = np.loadtxt(freq_filename, dtype=float)
#     lambdaSqArr_m2 = np.power(C / freqArr_Hz, 2.0)

#     phiArr_radm2 = fits_make_lin_axis(head, axis=FD_axis - 1)

#     HDUfwhm = pf.open(
#         FDF_filename.replace("FDF_tot_dirty", "RMSF_FWHM")
#         .replace("FDF_real_dirty", "RMSF_FWHM")
#         .replace("FDF_im_dirty", "RMSF_FWHM")
#         .replace("FDF_clean_tot", "RMSF_FWHM")
#         .replace("FDF_clean_real", "RMSF_FWHM")
#         .replace("FDF_clean_im", "RMSF_FWHM"),
#         "readonly",
#         memmap=True,
#     )
#     fwhmRMSF = HDUfwhm[0].data.squeeze()

#     return complex_cube, phiArr_radm2, fwhmRMSF, lambdaSqArr_m2, lam0Sq, head


# def main():
#     """
#     Start peak fitting from the command line.
#     """

#     import argparse

#     descStr = """
#     Perform pixel-wise fitting of the brightest peak in the FDF. This script
#     performs the same fitting used in the RMsynth_1D for every pixel in a 3D
#     FDF cube. The result is a set of 2D maps giving the fitting results (one
#     map per fitting output). These are saved to a set of FITS files.
#     """

#     epilog_text = """
#     The script saves 2D maps of the parameters from the peak fitting and
#     FDF characterization function. The user can select how many products are
#     saved:
#     no flag: a curated list of potentially interesting parameters is saved
#          -a: all outputs are saved.
#          -p: only parameters that describe the peak are saved.
#     """

#     # Parse the command line options
#     parser = argparse.ArgumentParser(
#         description=descStr,
#         epilog=epilog_text,
#         formatter_class=argparse.RawTextHelpFormatter,
#     )
#     parser.add_argument(
#         "FDF_filename",
#         metavar="FDF_file.fits",
#         nargs=1,
#         help="An FDF cube from RMsynth3D or RMclean3D",
#     )
#     parser.add_argument(
#         "freq_file",
#         metavar="freq_file.txt",
#         nargs=1,
#         help="A text file containing the channel frequencies.",
#     )
#     parser.add_argument(
#         "output_name",
#         metavar="output_name",
#         nargs=1,
#         help="The path and base name of the output maps.",
#     )
#     parser.add_argument(
#         "-a",
#         dest="all_products",
#         action="store_true",
#         help="Save all fitting products.",
#     )
#     parser.add_argument(
#         "-p", dest="peak_only", action="store_true", help="Save peak parameters only."
#     )
#     parser.add_argument(
#         "-v", dest="verbose", action="store_true", help="Verbose [False]."
#     )
#     parser.add_argument(
#         "-i",
#         dest="fitsI",
#         default=None,
#         help="FITS cube containing Stokes I model [None].",
#     )
#     parser.add_argument(
#         "-n",
#         dest="noiseFile",
#         default=None,
#         help="FITS file or cube containing noise values [None].",
#     )
#     parser.add_argument(
#         "-w",
#         dest="weightType",
#         default="uniform",
#         help="weighting ['uniform'] (all 1s) or 'variance' used in rmsynth3d, affects uncertainty estimation.",
#     )

#     args = parser.parse_args()

#     # Check for incompatible options:
#     if args.peak_only and args.all_products:
#         print("Please select at most ONE of the -a and -p flags.")
#         sys.exit()
#     if not os.path.exists(args.FDF_filename[0]):
#         print("Cannot find FDF file. Please check filename/path.")
#         sys.exit()
#     if not os.path.exists(args.freq_file[0]):
#         print("Cannot find frequency file. Please check filename/path.")
#         sys.exit()

#     if args.output_name[0][0] != "/":
#         args.output_name[0] = "./" + args.output_name[0]

#     # Define default product lists:
#     if args.peak_only:
#         product_list = [
#             "dFDFcorMAD",
#             "phiPeakPIfit_rm2",
#             "dPhiPeakPIfit_rm2",
#             "ampPeakPIfitEff",
#             "dAmpPeakPIfit",
#             "snrPIfit",
#             "polAngle0Fit_deg",
#             "dPolAngle0Fit_deg",
#         ]
#     elif args.all_products:
#         product_list = [
#             "dFDFcorMAD",
#             "phiPeakPIfit_rm2",
#             "dPhiPeakPIfit_rm2",
#             "ampPeakPIfit",
#             "ampPeakPIfitEff",
#             "dAmpPeakPIfit",
#             "snrPIfit",
#             "indxPeakPIfit",
#             "peakFDFimagFit",
#             "peakFDFrealFit",
#             "polAngleFit_deg",
#             "dPolAngleFit_deg",
#             "polAngle0Fit_deg",
#             "dPolAngle0Fit_deg",
#             "Ifreq0",
#             "dFDFth",
#             "lam0Sq_m2",
#             "freq0_Hz",
#             "fwhmRMSF",
#             "min_freq",
#             "max_freq",
#             "N_channels",
#             "median_channel_width",
#             "fracPol",
#         ]
#     else:  # Default option is a curated list of products I think are most useful.
#         product_list = [
#             "dFDFcorMAD",
#             "phiPeakPIfit_rm2",
#             "dPhiPeakPIfit_rm2",
#             "ampPeakPIfitEff",
#             "dAmpPeakPIfit",
#             "snrPIfit",
#             "peakFDFimagFit",
#             "peakFDFrealFit",
#             "Ifreq0",
#             "lam0Sq_m2",
#             "polAngle0Fit_deg",
#             "dPolAngle0Fit_deg",
#         ]

#     # Read in files
#     FDF, phiArr_radm2, fwhmRMSF, lambdaSqArr_m2, lam0Sq, header = read_files(
#         args.FDF_filename[0], args.freq_file[0]
#     )

#     if args.fitsI is not None:
#         dataI = readFitsCube(args.fitsI, args.verbose)[1]
#     else:
#         dataI = None
#     if args.noiseFile is not None:
#         rmsArr = readFreqFile(args.noiseFile, args.verbose)
#     else:
#         rmsArr = None

#     # Fit peaks
#     map_dict = pixelwise_peak_fitting(
#         FDF,
#         phiArr_radm2,
#         fwhmRMSF,
#         lambdaSqArr_m2,
#         lam0Sq,
#         product_list,
#         noiseArr=rmsArr,
#         stokesIcube=dataI,
#         weightType=args.weightType,
#     )

#     save_maps(map_dict, args.output_name[0], header)


# if __name__ == "__main__":
#     main()



"""RM-ms-clean utils"""

from __future__ import annotations

import logging
from functools import partial
from typing import Callable, Literal

import numpy as np
from numpy.typing import NDArray
from scipy.ndimage import convolve
from tqdm.auto import tqdm

from rm_lite.utils.clean import (
    CleanLoopResults,
    MinorLoopResults,
    RMCleanResults,
    minor_loop,
    restore_fdf,
)
from rm_lite.utils.fitting import (
    fit_rmsf,
    fwhm_to_sigma,
    unit_centred_gaussian,
)
from rm_lite.utils.logging import TqdmToLogger, logger
from rm_lite.utils.synthesis import compute_rmsf_params

TQDM_OUT = TqdmToLogger(logger, level=logging.INFO)

MSG = "This module is not yet implemented."
raise NotImplementedError(MSG)


@np.vectorize
def _scale_bias_function(
    scale: float,
    scale_0: float,
    scale_bias: float,
) -> float:
    """Offringa et al. (2017) scale-bias function.

    Args:
        scale (float): Scale parameter (relative to PSF FWHM)
        scale_0 (float): The first non-zero scale parameter
        scale_bias (float): The scale-bias parameter

    Returns:
        float: Weighting factor per scale
    """
    if scale == 0:
        return 1.0
    return scale_bias ** -(-1 - np.log2(scale / scale_0))


def scale_bias_function(
    scales: NDArray[np.float64],
    scale_bias: float,
) -> NDArray[np.float64]:
    """Offringa et al. (2017) scale-bias function.

    Args:
        scales (NDArray[np.float64]): Scale parameters (relative to PSF FWHM)
        scale_bias (float): The scale-bias parameter

    Returns:
        NDArray[np.float64]: Weighting factors per scale
    """
    if len(scales) == 1:
        return np.ones_like(scales)
    return _scale_bias_function(scales, scale_0=scales[1], scale_bias=scale_bias)


def scale_bias_function_cornwell(scales: NDArray[np.float64]) -> NDArray[np.float64]:
    return 1 - 0.6 * scales / scales.max()


def hanning(x_arr: NDArray[np.float64], length: float) -> NDArray[np.float64]:
    """Hanning window function.

    Args:
        x_arr (NDArray[np.float64]): Array of x values
        length (float): Length of the window

    Returns:
        NDArray[np.float64]: Hanning window function array
    """
    han = (1 / length) * np.cos(np.pi * x_arr / length) ** 2
    return np.where(np.abs(x_arr) < length / 2, han, 0)


def tapered_quad_kernel_function(
    phi_double_arr_radm2: NDArray[np.float64],
    scale: float,
    rmsf_fwhm: float,
    sum_normalised: bool = True,
) -> NDArray[np.float64]:
    """Tapered quadratic kernel function.

    Args:
        phi_double_arr_radm2 (NDArray[np.float64]): Phi array in rad/m^2
        scale (float): Scale (in FWHM units)
        rmsf_fwhm (float): RMSF FWHM in rad/m^2

    Returns:
        NDArray[np.float64]: Kernel function array (sum normalised)
    """
    scale_radm2 = scale * rmsf_fwhm
    kernel = hanning(phi_double_arr_radm2, scale_radm2) * (
        1 - (np.abs(phi_double_arr_radm2) / scale_radm2) ** 2
    )
    if sum_normalised:
        kernel /= kernel.sum()
    else:
        kernel /= kernel.max()
    return kernel


def gaussian_scale_kernel_function(
    phi_double_arr_radm2: NDArray[np.float64],
    scale: float,
    rmsf_fwhm: float,
    sum_normalised: bool = True,
) -> NDArray[np.float64]:
    """Gaussian scale kernel function.

    Args:
        phi_double_arr_radm2 (NDArray[np.float64]): Phi array in rad/m^2
        scale (float): Scale (in FWHM units)
        rmsf_fwhm (float): RMSF FWHM in rad/m^2

    Returns:
        NDArray[np.float64]: Kernel function array (sum normalised)
    """
    rmsf_sigma = fwhm_to_sigma(rmsf_fwhm)
    sigma = (3 / 16) * scale * rmsf_sigma
    kernel = unit_centred_gaussian(
        x=phi_double_arr_radm2,
        stddev=sigma,
    )
    if sum_normalised:
        kernel /= kernel.sum()
    else:
        kernel /= kernel.max()
    return kernel


KERNEL_FUNCS: dict[str, Callable] = {
    "tapered_quad": tapered_quad_kernel_function,
    "gaussian": gaussian_scale_kernel_function,
}


def convolve_fdf_scale(
    scale: float,
    fwhm: float,
    fdf_arr: NDArray[np.float64],
    phi_double_arr_radm2: NDArray[np.float64],
    kernel: Literal["tapered_quad", "gaussian"] = "gaussian",
    sum_normalised: bool = True,
) -> NDArray[np.float64]:
    """Convolve the FDF with a Gaussian kernel.

    Args:
        scale (float): Scale parameter (relative to PSF FWHM)
        fwhm (float): FWHM of the RMSF
        fdf_arr (NDArray[np.float64]): FDF array (complex)
        phi_double_arr_radm2 (NDArray[np.float64]): Double-length Faraday depth array (rad/m^2)
        kernel (Literal["tapered_quad", "gaussian"]): Kernel function

    Raises:
        ValueError: If an invalid normalization method is provided

    Returns:
        NDArray[np.float64]: Convolved FDF array
    """
    if scale == 0:
        return fdf_arr
    kernel_func = KERNEL_FUNCS.get(kernel, gaussian_scale_kernel_function)
    kernel_func_partial = partial(kernel_func, sum_normalised=sum_normalised)
    kernel_arr = kernel_func_partial(phi_double_arr_radm2, scale, fwhm)

    if np.iscomplexobj(fdf_arr):
        kernel_arr = kernel_arr * (1 + 1j) / np.sqrt(2)

    mode = "reflect"
    conv_spec = convolve(fdf_arr, kernel_arr, mode=mode)
    if mode == "valid":
        conv_spec = conv_spec[1:-1]

    assert len(conv_spec) == len(fdf_arr), "Convolved FDF has wrong length."

    return conv_spec


def find_significant_scale(
    scales: NDArray[np.float64],
    scale_bias: float,
    fdf_arr: NDArray[np.float64],
    phi_double_arr_radm2: NDArray[np.float64],
    fwhm: float,
    kernel: Literal["tapered_quad", "gaussian"] = "gaussian",
) -> tuple[float, float]:
    scale_parameters = scale_bias_function(scales, scale_bias)
    # scale_parameters = scale_bias_function_cornwell(scales)
    peaks = np.zeros_like(scales)
    for i, scale in enumerate(scales):
        fdf_conv = convolve_fdf_scale(
            scale=scale,
            fdf_arr=fdf_arr,
            fwhm=fwhm,
            phi_double_arr_radm2=phi_double_arr_radm2,
            kernel=kernel,
            sum_normalised=False,
        )
        peak = np.max(np.abs(fdf_conv))
        peaks[i] = peak
    scale_norm = scales.copy()
    scale_norm[scales == 0] = 1
    activated_index = np.argmax(peaks * scale_parameters / scale_norm)
    return scales[activated_index], scale_parameters[activated_index]


def multiscale_minor_loop(
    scales: NDArray[np.float64],
    scale_bias: float,
    resid_fdf_spectrum_mask: np.ma.MaskedArray,
    phi_arr_radm2: NDArray[np.float64],
    phi_double_arr_radm2: NDArray[np.float64],
    rmsf_spectrum: NDArray[np.float64],
    rmsf_fwhm: float,
    max_iter: int,
    max_iter_sub_minor: int,
    gain: float,
    mask: float,
    threshold: float,
    start_iter: int = 0,
    update_mask: bool = True,
    kernel: Literal["tapered_quad", "gaussian"] = "gaussian",
) -> MinorLoopResults:
    # Trust nothing
    resid_fdf_spectrum_mask = resid_fdf_spectrum_mask.copy()
    resid_fdf_spectrum = resid_fdf_spectrum_mask.data.copy()
    model_fdf_spectrum = np.zeros_like(resid_fdf_spectrum)
    clean_fdf_spectrum = np.zeros_like(resid_fdf_spectrum)
    rmsf_spectrum = rmsf_spectrum.copy()
    phi_arr_radm2 = phi_arr_radm2.copy()
    mask_arr = ~resid_fdf_spectrum_mask.mask.copy()
    iter_count = start_iter

    logger.info(f"Starting multiscale cycles...cleaning {mask_arr.sum()} pixels")
    for iter_count in range(start_iter, max_iter + 1):
        # Break conditions
        if resid_fdf_spectrum_mask.mask.all():
            logger.warning(
                f"All channels masked. Exiting loop...performed {iter_count} M-S iterations"
            )
            break
        if iter_count == max_iter:
            logger.warning(
                f"Max iterations reached. Exiting loop...performed {iter_count} M-S iterations"
            )
            break
        if np.ma.max(np.ma.abs(resid_fdf_spectrum_mask)) < threshold:
            logger.info(
                f"Threshold reached. Exiting loop...performed {iter_count} M-S iterations"
            )
            break

        activated_scale, scale_parameter = find_significant_scale(
            scales=scales,
            scale_bias=scale_bias,
            fdf_arr=np.abs(resid_fdf_spectrum),
            fwhm=rmsf_fwhm,
            phi_double_arr_radm2=phi_double_arr_radm2,
        )
        # activated_scale = 20
        logger.info(f"Cleaning activated scale: {activated_scale}")
        if activated_scale == 0:
            resid_fdf_spectrum_conv = resid_fdf_spectrum.copy()
            # peak_find_arr = np.abs(resid_fdf_spectrum)
        else:
            resid_fdf_spectrum_conv_abs = convolve_fdf_scale(
                scale=activated_scale,
                fdf_arr=np.abs(resid_fdf_spectrum),
                fwhm=rmsf_fwhm,
                phi_double_arr_radm2=phi_double_arr_radm2,
                kernel=kernel,
                sum_normalised=True,
            )
            resid_fdf_spectrum_conv = resid_fdf_spectrum_conv_abs * np.exp(
                1j * np.angle(resid_fdf_spectrum)
            )
            # peak_find_arr = convolve_fdf_scale(
            #     scale=activated_scale,
            #     fdf_arr=np.abs(resid_fdf_spectrum),
            #     fwhm=rmsf_fwhm,
            #     phi_double_arr_radm2=phi_double_arr_radm2,
            #     kernel=kernel,
            # )
        # resid_fdf_spectrum_conv *= scale_parameter

        if activated_scale == 0:
            rmsf_spectrum_conv = rmsf_spectrum.copy()
        else:
            rmsf_spectrum_conv_abs = convolve_fdf_scale(
                scale=activated_scale,
                fdf_arr=np.abs(rmsf_spectrum),
                fwhm=rmsf_fwhm,
                phi_double_arr_radm2=phi_double_arr_radm2,
                kernel=kernel,
                sum_normalised=True,
            )
            rmsf_spectrum_conv = rmsf_spectrum_conv_abs * np.exp(
                1j * np.angle(rmsf_spectrum)
            )

        scale_factor = np.nanmax(np.abs(rmsf_spectrum_conv))
        rmsf_spectrum_conv /= scale_factor
        resid_fdf_spectrum_conv /= scale_factor

        # Redo
        # rmsf_spectrum_conv = convolve_fdf_scale(
        #     scale=activated_scale,
        #     fdf_arr=rmsf_spectrum,
        #     fwhm=rmsf_fwhm,
        #     phi_double_arr_radm2=phi_double_arr_radm2,
        #     kernel=kernel,
        # )
        # rmsf_spectrum_conv /= np.nanmax(np.abs(rmsf_spectrum_conv))

        if update_mask:
            mask_arr = np.abs(resid_fdf_spectrum_conv) > mask
        resid_fdf_spectrum_mask_conv = np.ma.array(
            resid_fdf_spectrum_conv, mask=~mask_arr
        )

        conv_fwhm = fit_rmsf(
            np.abs(rmsf_spectrum_conv),
            phi_double_arr_radm2=phi_double_arr_radm2,
            fwhm_rmsf_radm2=rmsf_fwhm * activated_scale,
        )
        sub_minor_results = minor_loop(
            resid_fdf_spectrum_mask=resid_fdf_spectrum_mask_conv,
            phi_arr_radm2=phi_arr_radm2,
            phi_double_arr_radm2=phi_double_arr_radm2,
            rmsf_spectrum=rmsf_spectrum_conv,
            rmsf_fwhm=float(conv_fwhm),
            # rmsf_fwhm=rmsf_fwhm * activated_scale,
            max_iter=max_iter_sub_minor,
            gain=gain,
            mask=mask * activated_scale,
            threshold=threshold,
            start_iter=iter_count,
            update_mask=update_mask,
            # peak_find_arr=peak_find_arr,
        )

        # Convolve the clean components with the RMSF
        clean_deltas = sub_minor_results.model_fdf_spectrum
        iter_count += sub_minor_results.iter_count  # noqa: PLW2901
        if activated_scale == 0:
            clean_model = clean_deltas
        else:
            clean_model = convolve_fdf_scale(
                scale=activated_scale,
                fwhm=rmsf_fwhm,
                fdf_arr=clean_deltas,
                phi_double_arr_radm2=phi_double_arr_radm2,
                kernel=kernel,
                sum_normalised=True,
            )
        model_fdf_spectrum += clean_model

        clean_spectrum = restore_fdf(
            model_fdf_spectrum=model_fdf_spectrum,
            phi_double_arr_radm2=phi_double_arr_radm2,
            fwhm_rmsf=rmsf_fwhm,
        )
        shifted_rmsf = sub_minor_results.model_rmsf_spectrum
        # shifted_rmsf = np.convolve(
        #     clean_model,
        #     rmsf_spectrum / gaussian_integrand(1, fwhm=rmsf_fwhm),
        #     mode="valid",
        # )[1:-1]

        clean_fdf_spectrum += clean_spectrum
        resid_fdf_spectrum -= shifted_rmsf

        if update_mask:
            mask_arr = np.abs(resid_fdf_spectrum) > mask

        resid_fdf_spectrum_mask = np.ma.array(resid_fdf_spectrum, mask=~mask_arr)

    return MinorLoopResults(
        clean_fdf_spectrum=clean_fdf_spectrum,
        resid_fdf_spectrum=resid_fdf_spectrum,
        resid_fdf_spectrum_mask=resid_fdf_spectrum_mask,
        model_fdf_spectrum=model_fdf_spectrum,
        model_rmsf_spectrum=np.zeros_like(model_fdf_spectrum),
        iter_count=iter_count,
    )


def multiscale_cycle(
    scales: NDArray[np.float64],
    scale_bias: float,
    phi_arr_radm2: NDArray[np.float64],
    phi_double_arr_radm2: NDArray[np.float64],
    dirty_fdf_spectrum: NDArray[np.float64],
    rmsf_spectrum: NDArray[np.float64],
    rmsf_fwhm: float,
    mask: float,
    threshold: float,
    max_iter: int,
    max_iter_sub_minor: int,
    gain: float,
    kernel: Literal["tapered_quad", "gaussian"] = "gaussian",
) -> CleanLoopResults:
    # Initialise arrays to hold the residual FDF, clean components, clean FDF
    resid_fdf_spectrum = dirty_fdf_spectrum.copy()
    mask_arr = np.abs(resid_fdf_spectrum) > mask
    resid_fdf_spectrum_mask = np.ma.array(resid_fdf_spectrum, mask=~mask_arr)

    initial_loop_results = multiscale_minor_loop(
        scales=scales,
        scale_bias=scale_bias,
        resid_fdf_spectrum_mask=resid_fdf_spectrum_mask,
        phi_arr_radm2=phi_arr_radm2,
        phi_double_arr_radm2=phi_double_arr_radm2,
        rmsf_spectrum=rmsf_spectrum,
        rmsf_fwhm=rmsf_fwhm,
        max_iter_sub_minor=max_iter_sub_minor,
        max_iter=max_iter,
        gain=gain,
        mask=mask,
        threshold=threshold,
        start_iter=0,
        update_mask=True,
        kernel=kernel,
    )

    # Deep clean
    # Mask where clean components have been added
    mask_arr = np.abs(initial_loop_results.model_fdf_spectrum) > 0
    resid_fdf_spectrum = initial_loop_results.resid_fdf_spectrum
    resid_fdf_spectrum_mask = np.ma.array(resid_fdf_spectrum, mask=~mask_arr)

    logger.info(f"Starting deep clean...cleaning {mask_arr.sum()} pixels")
    deep_loop_results = multiscale_minor_loop(
        scales=scales,
        scale_bias=scale_bias,
        resid_fdf_spectrum_mask=resid_fdf_spectrum_mask,
        phi_arr_radm2=phi_arr_radm2,
        phi_double_arr_radm2=phi_double_arr_radm2,
        rmsf_spectrum=rmsf_spectrum,
        rmsf_fwhm=rmsf_fwhm,
        max_iter=max_iter,
        max_iter_sub_minor=max_iter_sub_minor,
        gain=gain,
        mask=mask,
        threshold=threshold,
        start_iter=0,
        update_mask=False,
        kernel=kernel,
    )

    clean_fdf_spectrum = np.squeeze(
        deep_loop_results.clean_fdf_spectrum + initial_loop_results.clean_fdf_spectrum
    )
    resid_fdf_spectrum = np.squeeze(deep_loop_results.resid_fdf_spectrum)
    model_fdf_spectrum = np.squeeze(
        deep_loop_results.model_fdf_spectrum + initial_loop_results.model_fdf_spectrum
    )

    return CleanLoopResults(
        clean_fdf_spectrum=clean_fdf_spectrum,
        resid_fdf_spectrum=resid_fdf_spectrum,
        model_fdf_spectrum=model_fdf_spectrum,
        iter_count=deep_loop_results.iter_count,
    )


def mutliscale_rmclean(
    freq_arr_hz: NDArray[np.float64],
    dirty_fdf_arr: NDArray[np.float64],
    phi_arr_radm2: NDArray[np.float64],
    rmsf_arr: NDArray[np.float64],
    phi_double_arr_radm2: NDArray[np.float64],
    fwhm_rmsf_arr: NDArray[np.float64],
    mask: float,
    threshold: float,
    max_iter: int = 1000,
    max_iter_sub_minor: int = 10_000,
    gain: float = 0.1,
    scale_bias: float = 0.9,
    scales: NDArray[np.float64] | None = None,
    mask_arr: NDArray[np.float64] | None = None,
    kernel: Literal["tapered_quad", "gaussian"] = "gaussian",
) -> RMCleanResults:
    _bad_result = RMCleanResults(
        clean_fdf_arr=dirty_fdf_arr,
        model_fdf_arr=np.zeros_like(dirty_fdf_arr),
        clean_iter_arr=np.zeros_like(phi_arr_radm2),
        resid_fdf_arr=dirty_fdf_arr,
    )
    # Sanity checks on array sizes
    n_phi = phi_arr_radm2.shape[0]
    if n_phi != dirty_fdf_arr.shape[0]:
        logger.error("'phi_arr_radm2' and 'dirty_fdf_arr' are not the same length.")
        return _bad_result
    n_phi2 = phi_double_arr_radm2.shape[0]
    if n_phi2 != rmsf_arr.shape[0]:
        logger.error("mismatch in 'phi_double_arr_radm2' and 'rmsf_arr' length.")
        return _bad_result
    if not (n_phi2 >= 2 * n_phi):
        logger.error("the Faraday depth of the RMSF must be twice the FDF.")
        return _bad_result
    n_dimension = len(dirty_fdf_arr.shape)
    if not n_dimension <= 3:
        logger.error("FDF array dimensions must be <= 3.")
        return _bad_result
    if n_dimension != len(rmsf_arr.shape):
        logger.error("the input RMSF and FDF must have the same number of axes.")
        return _bad_result
    if rmsf_arr.shape[1:] != dirty_fdf_arr.shape[1:]:
        logger.error("the xy dimensions of the RMSF and FDF must match.")
        return _bad_result
    if mask_arr is not None:
        if mask_arr.shape != dirty_fdf_arr.shape[1:]:
            logger.error("pixel mask must match xy dimension of FDF cube.")
            return _bad_result
    else:
        mask_arr = np.ones(dirty_fdf_arr.shape[1:], dtype=bool)

    # Reshape the FDF & RMSF array to 3 dimensions and mask array to 2
    if n_dimension == 1:
        dirty_fdf_arr = np.reshape(dirty_fdf_arr, (dirty_fdf_arr.shape[0], 1, 1))
        rmsf_arr = np.reshape(rmsf_arr, (rmsf_arr.shape[0], 1, 1))
        mask_arr = np.reshape(mask_arr, (1, 1))
        fwhm_rmsf_arr = np.reshape(fwhm_rmsf_arr, (1, 1))
    elif n_dimension == 2:
        dirty_fdf_arr = np.reshape(dirty_fdf_arr, [*list(dirty_fdf_arr.shape[:2]), 1])
        rmsf_arr = np.reshape(rmsf_arr, [*list(rmsf_arr.shape[:2]), 1])
        mask_arr = np.reshape(mask_arr, (dirty_fdf_arr.shape[1], 1))
        fwhm_rmsf_arr = np.reshape(fwhm_rmsf_arr, (dirty_fdf_arr.shape[1], 1))

    # Compute the scale parameters
    rmsf_params = compute_rmsf_params(
        freq_arr_hz=freq_arr_hz,
        weight_arr=np.ones_like(freq_arr_hz),
        super_resolution=False,
    )
    max_scale = rmsf_params.phi_max_scale / rmsf_params.rmsf_fwhm_meas
    logger.info(
        f"Maximum Faraday scale {rmsf_params.phi_max_scale:0.2f} / (rad/m^2) -- {max_scale:0.2f} / RMSF FWHM."
    )
    if scales is None:
        scales = np.arange(0, max_scale, step=0.1)

    logger.info(f"Using scales: {scales}")

    if scales.max() > max_scale:
        logger.warning(
            f"Maximum scale parameter {scales.max()} is greater than the RMSF max scale {max_scale}."
        )
    iter_count_arr = np.zeros_like(mask_arr, dtype=int)

    # Determine which pixels have components above the cutoff
    abs_fdf_cube = np.abs(np.nan_to_num(dirty_fdf_arr))
    cutoff_mask = np.where(np.max(abs_fdf_cube, axis=0) >= mask, 1, 0)
    pixels_to_clean = np.rot90(np.where(cutoff_mask > 0))

    num_pixels = dirty_fdf_arr.shape[-1] * dirty_fdf_arr.shape[-2]
    num_pixels_clean = len(pixels_to_clean)
    logger.info(f"Cleaning {num_pixels_clean}/{num_pixels} spectra.")

    # Initialise arrays to hold the residual FDF, clean components, clean FDF
    # Residual is initially copies of dirty FDF, so that pixels that are not
    #  processed get correct values (but will be overridden when processed)
    clean_fdf_spectrum = np.zeros_like(dirty_fdf_arr)
    model_fdf_spectrum = np.zeros(dirty_fdf_arr.shape, dtype=complex)
    resid_fdf_arr = dirty_fdf_arr.copy()

    # Loop through the pixels containing a polarised signal
    for yi, xi in tqdm(pixels_to_clean):
        clean_loop_results = multiscale_cycle(
            scales=scales,
            scale_bias=scale_bias,
            phi_arr_radm2=phi_arr_radm2,
            phi_double_arr_radm2=phi_double_arr_radm2,
            dirty_fdf_spectrum=dirty_fdf_arr[:, yi, xi],
            rmsf_spectrum=rmsf_arr[:, yi, xi],
            rmsf_fwhm=fwhm_rmsf_arr[yi, xi],
            mask=mask,
            threshold=threshold,
            max_iter=max_iter,
            max_iter_sub_minor=max_iter_sub_minor,
            gain=gain,
            kernel=kernel,
        )
        clean_fdf_spectrum[:, yi, xi] = clean_loop_results.clean_fdf_spectrum
        resid_fdf_arr[:, yi, xi] = clean_loop_results.resid_fdf_spectrum
        model_fdf_spectrum[:, yi, xi] = clean_loop_results.model_fdf_spectrum
        iter_count_arr[yi, xi] = clean_loop_results.iter_count

    # Restore the residual to the cleaned FDF (moved outside of loop:
    # will now work for pixels/spectra without clean components)
    clean_fdf_spectrum += resid_fdf_arr

    # Remove redundant dimensions
    clean_fdf_spectrum = np.squeeze(clean_fdf_spectrum)
    model_fdf_spectrum = np.squeeze(model_fdf_spectrum)
    iter_count_arr = np.squeeze(iter_count_arr)
    resid_fdf_arr = np.squeeze(resid_fdf_arr)

    return RMCleanResults(
        clean_fdf_spectrum, model_fdf_spectrum, iter_count_arr, resid_fdf_arr
    )



"""Tests for the RM synthesis and related tools"""

from __future__ import annotations

from typing import NamedTuple

import numpy as np
import pytest
from numpy.typing import NDArray
from rm_lite.tools_1d.rmsynth import run_rmsynth
from rm_lite.utils.logging import logger
from rm_lite.utils.synthesis import (
    FWHM,
    freq_to_lambda2,
    get_fwhm_rmsf,
    lambda2_to_freq,
    make_phi_arr,
    rmsynth_nufft,
)

RNG = np.random.default_rng()


class MockData(NamedTuple):
    freqs: NDArray[np.float64]
    lsq: NDArray[np.float64]
    stokes_i: NDArray[np.float64]
    stokes_q: NDArray[np.float64]
    stokes_u: NDArray[np.float64]


class MockModel(NamedTuple):
    flux: float
    frac_pol: float
    rm: float
    pa_0: float
    fwhm: float


@pytest.fixture
def racs_model() -> MockModel:
    fwhm = 49.57
    rm = RNG.uniform(-1000, 1000)
    pa = RNG.uniform(0, 180)
    frac_pol = RNG.uniform(0.5, 0.7)
    flux = RNG.uniform(1, 10)

    return MockModel(flux, frac_pol, rm, pa, fwhm)


@pytest.fixture
def racs_data(racs_model):
    freqs = np.arange(744, 1032, 1) * 1e6
    lsq = freq_to_lambda2(freqs)
    stokes_i = np.ones_like(freqs) * racs_model.flux
    stokes_q = (
        stokes_i
        * racs_model.frac_pol
        * np.cos(2 * racs_model.rm * lsq + 2 * racs_model.pa_0)
    )
    stokes_u = (
        stokes_i
        * racs_model.frac_pol
        * np.sin(2 * racs_model.rm * lsq + 2 * racs_model.pa_0)
    )
    return MockData(freqs, lsq, stokes_i, stokes_q, stokes_u)


def test_get_fwhm_rmsf(racs_data, racs_model):
    assert np.allclose(racs_data.lsq, freq_to_lambda2(lambda2_to_freq(racs_data.lsq)))
    fwhm: FWHM = get_fwhm_rmsf(racs_data.lsq)
    assert np.isclose(fwhm.fwhm_rmsf_radm2, racs_model.fwhm, atol=1)
    assert np.isclose(
        fwhm.d_lambda_sq_max_m2, np.nanmax(np.abs(np.diff(racs_data.lsq)))
    )
    assert np.isclose(
        fwhm.lambda_sq_range_m2,
        np.nanmax(racs_data.lsq) - np.nanmin(racs_data.lsq),
    )


def test_rmsynth_nufft(racs_data: MockData, racs_model: MockModel):
    phis = make_phi_arr(
        phi_max_radm2=1000,
        d_phi_radm2=1,
    )
    fdf_dirty = rmsynth_nufft(
        complex_pol_arr=racs_data.stokes_q + 1j * racs_data.stokes_u,
        lambda_sq_arr_m2=racs_data.lsq,
        phi_arr_radm2=phis,
        weight_arr=np.ones_like(racs_data.stokes_q),
        lam_sq_0_m2=float(np.mean(racs_data.lsq)),
    )

    peak_rm = phis[np.argmax(np.abs(fdf_dirty))]
    assert np.isclose(peak_rm, racs_model.rm, atol=1)


def test_run_rmsynth(racs_data: MockData, racs_model: MockModel):
    complex_data = racs_data.stokes_q + 1j * racs_data.stokes_u
    complex_error = np.ones_like(racs_data.stokes_q) + 1j * np.ones_like(
        racs_data.stokes_u
    )
    complex_error *= 1e-3

    rmsyth_results = run_rmsynth(
        freq_arr_hz=racs_data.freqs,
        complex_pol_arr=complex_data,
        complex_pol_error=complex_error,
        stokes_i_arr=racs_data.stokes_i,
        stokes_i_error_arr=np.ones_like(racs_data.stokes_i) * 1e-3,
    )

    fdf_parameters = rmsyth_results.fdf_parameters
    logger.info(fdf_parameters)

    assert np.isclose(
        fdf_parameters["peak_rm_fit"][0],
        racs_model.rm,
        # atol=fdf_parameters["peak_rm_fit_error"][0],
        atol=1,
    )

    assert np.isclose(
        fdf_parameters["frac_pol"].to_numpy()[0],
        racs_model.frac_pol,
        # atol=fdf_parameters["frac_pol_error"].to_numpy()[0],
        atol=0.1,
    )


def test_2d_synth(racs_data: MockData, racs_model: MockModel):
    stokes_q = racs_data.stokes_q
    stokes_u = racs_data.stokes_u
    complex_pol_arr = stokes_q + 1j * stokes_u
    complesx_pol_2d = np.tile(complex_pol_arr, (10, 1))

    phis = make_phi_arr(
        phi_max_radm2=1000,
        d_phi_radm2=1,
    )
    weights = np.ones_like(stokes_q)
    lambda_0_m2 = float(np.mean(racs_data.lsq))

    with pytest.raises(
        ValueError,
        match=r"Data depth does not match lambda\^2 vector \((\d+) vs (\d+)\)\.",
    ):
        dirty_fdf = rmsynth_nufft(
            complex_pol_arr=complesx_pol_2d,
            lambda_sq_arr_m2=racs_data.lsq,
            phi_arr_radm2=phis,
            weight_arr=weights,
            lam_sq_0_m2=lambda_0_m2,
        )

    dirty_fdf = rmsynth_nufft(
        complex_pol_arr=complesx_pol_2d.T,
        lambda_sq_arr_m2=racs_data.lsq,
        phi_arr_radm2=phis,
        weight_arr=weights,
        lam_sq_0_m2=lambda_0_m2,
    )

    peak_rms = phis[:, np.newaxis][np.argmax(np.abs(dirty_fdf), axis=0)].squeeze()
    peak_pis = np.max(np.abs(dirty_fdf), axis=0)
    assert np.allclose(peak_rms, racs_model.rm, atol=1)
    assert np.allclose(peak_pis, racs_model.frac_pol * racs_model.flux, atol=0.1)



"""RM-synthesis on 1D data"""

from __future__ import annotations

import time
from typing import Literal, NamedTuple

import numpy as np
import polars as pl
from numpy.typing import NDArray
from scipy import interpolate

from rm_lite.utils.logging import logger
from rm_lite.utils.synthesis import (
    FDFOptions,
    StokesData,
    compute_rmsynth_params,
    compute_theoretical_noise,
    create_fractional_spectra,
    get_fdf_parameters,
    get_rmsf_nufft,
    lambda2_to_freq,
    rmsynth_nufft,
)


class RMSynth1DResults(NamedTuple):
    """Resulting arrays from RM-synthesis"""

    fdf_parameters: pl.DataFrame
    """ FDF parameters """
    fdf_arrs: pl.DataFrame
    """ RMSynth arrays """
    rmsf_arrs: pl.DataFrame
    """ RMSF arrays """
    stokes_i_arrs: pl.DataFrame
    """ Stokes I arrays """


rmsyth_arrs_schema = pl.Schema(
    {
        "phi_arr_radm2": pl.Float64,
        "fdf_dirty_complex_arr": pl.Object,
    }
)
rmsyth_arrs_schema_df = rmsyth_arrs_schema.to_frame(eager=True)
rmsf_arrs_schema = pl.Schema(
    {
        "phi2_arr_radm2": pl.Float64,
        "rmsf_complex_arr": pl.Object,
    }
)
rmsf_arrs_schema_df = rmsf_arrs_schema.to_frame(eager=True)
stokes_i_arrs_schema = pl.Schema(
    {
        "freq_arr_hz": pl.Float64,
        "lambda_sq_arr_m2": pl.Float64,
        "stokes_i_model_arr": pl.Float64,
        "stokes_i_model_error": pl.Float64,
        "flag_arr": pl.Boolean,
        "complex_frac_pol_arr": pl.Object,
        "complex_frac_pol_error": pl.Object,
    }
)
stokes_i_arrs_schema_df = stokes_i_arrs_schema.to_frame(eager=True)


def run_rmsynth(
    freq_arr_hz: NDArray[np.float64],
    complex_pol_arr: NDArray[np.complex128],
    complex_pol_error: NDArray[np.complex128],
    stokes_i_arr: NDArray[np.float64] | None = None,
    stokes_i_error_arr: NDArray[np.float64] | None = None,
    stokes_i_model_arr: NDArray[np.float64] | None = None,
    stokes_i_model_error: NDArray[np.float64] | None = None,
    phi_max_radm2: float | None = None,
    d_phi_radm2: float | None = None,
    n_samples: float | None = 10.0,
    weight_type: Literal["variance", "uniform"] = "variance",
    do_fit_rmsf: bool = False,
    do_fit_rmsf_real: bool = False,
    fit_function: Literal["log", "linear"] = "log",
    fit_order: int = 2,
) -> RMSynth1DResults:
    """Run RM-synthesis on 1D data

    Args:
        freq_arr_hz (NDArray[np.float64]): Frequencies in Hz
        complex_pol_arr (NDArray[np.complex128]): Complex polarisation values (Q + iU)
        complex_pol_error (NDArray[np.float64]): Complex polarisation errors (dQ + idU)
        stokes_i_arr (NDArray[np.float64] | None, optional): Total itensity values. Defaults to None.
        stokes_i_error_arr (NDArray[np.float64] | None, optional): Total intensity errors. Defaults to None.
        stokes_i_model_arr (NDArray[np.float64] | None, optional): Total intensity model array. Defaults to None.
        stokes_i_model_error (NDArray[np.float64] | None, optional): Total intensity model error. Defaults to None.
        phi_max_radm2 (float | None, optional): Maximum Faraday depth. Defaults to None.
        d_phi_radm2 (float | None, optional): Spacing in Faraday depth. Defaults to None.
        n_samples (float | None, optional): Number of samples across the RMSF. Defaults to 10.0.
        weight_type ("variance", "uniform", optional): Type of weighting. Defaults to "variance".
        do_fit_rmsf (bool, optional): Fit the RMSF main lobe. Defaults to False.
        do_fit_rmsf_real (bool, optional): The the real part of the RMSF. Defaults to False.
        fit_function ("log" | "linear", optional): _description_. Defaults to "log".
        fit_order (int, optional): Polynomial fit order. Defaults to 2. Negative values will iterate until the fit is good.

    Returns:
        RMSynth1DResults:
            fdf_parameters (pl.DataFrame): FDF parameters
            fdf_arrs (pl.DataFrame): RMSynth arrays
            rmsf_arrs (pl.DataFrame): RMSF arrays
    """
    stokes_data = StokesData(
        freq_arr_hz=freq_arr_hz,
        complex_pol_arr=complex_pol_arr,
        complex_pol_error=complex_pol_error,
        stokes_i_arr=stokes_i_arr,
        stokes_i_error_arr=stokes_i_error_arr,
        stokes_i_model_arr=stokes_i_model_arr,
        stokes_i_model_error=stokes_i_model_error,
    )

    fdf_options = FDFOptions(
        phi_max_radm2=phi_max_radm2,
        d_phi_radm2=d_phi_radm2,
        n_samples=n_samples,
        weight_type=weight_type,
        do_fit_rmsf=do_fit_rmsf,
        do_fit_rmsf_real=do_fit_rmsf_real,
    )

    return _run_rmsynth(stokes_data, fdf_options, fit_function, fit_order)


def _run_rmsynth(
    stokes_data: StokesData,
    fdf_options: FDFOptions,
    fit_function: Literal["log", "linear"] = "log",
    fit_order: int = 2,
) -> RMSynth1DResults:
    """Run RM-synthesis on 1D data with packed data

    Args:
        stokes_data (StokesData): Frequency-dependent polarisation data
        fdf_options (FDFOptions): RM-synthesis options
        fit_function ("log", "linear", optional): Type of function to fit. Defaults to "log".
        fit_order (int, optional): Polynomial fit order. Defaults to 2. Negative values will iterate until the fit is good.

    Returns:
        RMSynth1DResults:
            fdf_parameters (pl.DataFrame): FDF parameters
            fdf_arrs (pl.DataFrame): RMSynth arrays
            rmsf_arrs (pl.DataFrame): RMSF arrays
    """

    rmsynth_params = compute_rmsynth_params(
        freq_arr_hz=stokes_data.freq_arr_hz,
        complex_pol_arr=stokes_data.complex_pol_arr,
        complex_pol_error=stokes_data.complex_pol_error,
        fdf_options=fdf_options,
    )

    fractional_spectra = create_fractional_spectra(
        stokes_data=stokes_data,
        ref_freq_hz=lambda2_to_freq(rmsynth_params.lam_sq_0_m2),
        fit_order=fit_order,
        fit_function=fit_function,
    )

    # Compute after any fractional spectra have been created
    tick = time.time()

    # Perform RM-synthesis on the spectrum
    no_nan_idx = fractional_spectra.no_nan_idx
    fdf_dirty_arr = rmsynth_nufft(
        complex_pol_arr=fractional_spectra.complex_pol_frac_arr[no_nan_idx],
        lambda_sq_arr_m2=rmsynth_params.lambda_sq_arr_m2[no_nan_idx],
        phi_arr_radm2=rmsynth_params.phi_arr_radm2,
        weight_arr=rmsynth_params.weight_arr[no_nan_idx],
        lam_sq_0_m2=rmsynth_params.lam_sq_0_m2,
    )

    # Calculate the Rotation Measure Spread Function
    rmsf_result = get_rmsf_nufft(
        lambda_sq_arr_m2=rmsynth_params.lambda_sq_arr_m2,
        phi_arr_radm2=rmsynth_params.phi_arr_radm2,
        weight_arr=rmsynth_params.weight_arr,
        lam_sq_0_m2=rmsynth_params.lam_sq_0_m2,
        mask_arr=~no_nan_idx,
        do_fit_rmsf=fdf_options.do_fit_rmsf,
        do_fit_rmsf_real=fdf_options.do_fit_rmsf_real,
    )

    tock = time.time()
    cpu_time = tock - tick
    logger.info(f"RM-synthesis completed in {cpu_time * 1000:.2f}ms.")

    theoretical_noise = compute_theoretical_noise(
        complex_pol_frac_error=fractional_spectra.complex_pol_frac_error[no_nan_idx],
        weight_arr=rmsynth_params.weight_arr[no_nan_idx],
    )

    assert stokes_data.freq_arr_hz.shape == fractional_spectra.stokes_i_model_arr.shape
    stokes_i_model = interpolate.interp1d(
        stokes_data.freq_arr_hz[no_nan_idx],
        fractional_spectra.stokes_i_model_arr[no_nan_idx],
    )

    stokes_i_reference_flux = stokes_i_model(
        lambda2_to_freq(rmsynth_params.lam_sq_0_m2)
    )
    fdf_dirty_arr *= stokes_i_reference_flux
    theoretical_noise = theoretical_noise.with_options(  # type: ignore[no-untyped-call]
        fdf_error_noise=theoretical_noise.fdf_error_noise * stokes_i_reference_flux,
        fdf_q_noise=theoretical_noise.fdf_q_noise * stokes_i_reference_flux,
        fdf_u_noise=theoretical_noise.fdf_u_noise * stokes_i_reference_flux,
    )

    # Measure the parameters of the dirty FDF
    # Use the theoretical noise to calculate uncertainties
    fdf_parameters = get_fdf_parameters(
        fdf_arr=fdf_dirty_arr,
        phi_arr_radm2=rmsynth_params.phi_arr_radm2,
        fwhm_rmsf_radm2=float(rmsf_result.fwhm_rmsf_arr),
        freq_arr_hz=stokes_data.freq_arr_hz,
        complex_pol_arr=fractional_spectra.complex_pol_frac_arr,
        complex_pol_error=fractional_spectra.complex_pol_frac_error,
        lambda_sq_arr_m2=rmsynth_params.lambda_sq_arr_m2,
        lam_sq_0_m2=rmsynth_params.lam_sq_0_m2,
        stokes_i_reference_flux=stokes_i_reference_flux,
        theoretical_noise=theoretical_noise,
        fit_function=fit_function,
    )
    rmsyth_arrs = rmsyth_arrs_schema_df.vstack(
        pl.DataFrame(
            {
                "phi_arr_radm2": rmsynth_params.phi_arr_radm2,
                "fdf_dirty_complex_arr": fdf_dirty_arr,
            }
        )
    )

    rmsf_arrs = rmsf_arrs_schema_df.vstack(
        pl.DataFrame(
            {
                "phi2_arr_radm2": rmsf_result.phi_double_arr_radm2,
                "rmsf_complex_arr": rmsf_result.rmsf_cube,
            }
        )
    )
    stokes_i_arrs = stokes_i_arrs_schema_df.vstack(
        pl.DataFrame(
            {
                "freq_arr_hz": stokes_data.freq_arr_hz,
                "lambda_sq_arr_m2": rmsynth_params.lambda_sq_arr_m2,
                "stokes_i_model_arr": fractional_spectra.stokes_i_model_arr,
                "stokes_i_model_error": fractional_spectra.stokes_i_model_error,
                "flag_arr": no_nan_idx,
                "complex_frac_pol_arr": fractional_spectra.complex_pol_frac_arr,
                "complex_frac_pol_error": fractional_spectra.complex_pol_frac_error,
            }
        )
    )

    return RMSynth1DResults(fdf_parameters, rmsyth_arrs, rmsf_arrs, stokes_i_arrs)



"""RM-clean utils"""

from __future__ import annotations

import logging
from typing import NamedTuple, TypeVar

import numpy as np
from numpy.typing import NDArray
from tqdm.auto import tqdm

from rm_lite.utils.arrays import nd_to_two_d, two_d_to_nd
from rm_lite.utils.fitting import (
    gaussian,
    gaussian_integrand,
    unit_centred_gaussian,
)
from rm_lite.utils.logging import TqdmToLogger, logger

TQDM_OUT = TqdmToLogger(logger, level=logging.INFO)

DType = TypeVar("DType", bound=np.generic)


class RMCleanResults(NamedTuple):
    """Results of the RM-CLEAN calculation"""

    clean_fdf_arr: NDArray[np.complex128]
    """The cleaned Faraday dispersion function cube"""
    model_fdf_arr: NDArray[np.complex128]
    """The clean components cube"""
    clean_iter_arr: NDArray[np.int16]
    """The number of iterations for each pixel"""
    resid_fdf_arr: NDArray[np.complex128]
    """The residual Faraday dispersion function cube"""

    def with_options(self, **kwargs):
        as_dict = self._asdict()
        as_dict.update(kwargs)

        return RMCleanResults(**as_dict)


class CleanLoopResults(NamedTuple):
    """Results of the RM-CLEAN loop"""

    clean_fdf_spectrum: NDArray[np.complex128]
    """The cleaned Faraday dispersion function cube"""
    resid_fdf_spectrum: NDArray[np.complex128]
    """The residual Faraday dispersion function cube"""
    model_fdf_spectrum: NDArray[np.complex128]
    """The clean components cube"""
    iter_count: int
    """The number of iterations"""


class MinorLoopResults(NamedTuple):
    """Results of the RM-CLEAN minor loop"""

    clean_fdf_spectrum: NDArray[np.complex128]
    """The cleaned Faraday dispersion function cube"""
    resid_fdf_spectrum: NDArray[np.complex128]
    """The residual Faraday dispersion function cube"""
    resid_fdf_spectrum_mask: np.ma.MaskedArray
    """The masked residual Faraday dispersion function cube"""
    model_fdf_spectrum: NDArray[np.complex128]
    """The clean components cube"""
    model_rmsf_spectrum: NDArray[np.complex128]
    """ Model * RMSF """
    iter_count: int
    """The number of iterations"""


def restore_fdf(
    model_fdf_spectrum: NDArray[np.complex128],
    phi_double_arr_radm2: NDArray[np.float64],
    fwhm_rmsf: float,
) -> NDArray[np.complex128]:
    clean_beam = unit_centred_gaussian(
        x=phi_double_arr_radm2,
        fwhm=fwhm_rmsf,
    ) / gaussian_integrand(amplitude=1, fwhm=fwhm_rmsf)
    restored_fdf = np.convolve(
        model_fdf_spectrum.real, clean_beam, mode="valid"
    ) + 1j * np.convolve(model_fdf_spectrum.imag, clean_beam, mode="valid")
    return np.array(restored_fdf[1:-1])


class RMSynthArrays(NamedTuple):
    """Arrays for RM-synthesis"""

    dirty_fdf_arr: NDArray[np.complex128]
    """Dirty Faraday dispersion function array"""
    phi_arr_radm2: NDArray[np.float64]
    """Faraday depth array in rad/m^2"""
    phi_double_arr_radm2: NDArray[np.float64]
    """Double-length Faraday depth array in rad/m^2"""
    rmsf_arr: NDArray[np.complex128]
    """RMSF array"""
    fwhm_rmsf_arr: NDArray[np.float64]
    """FWHM of the RMSF array"""
    fdf_mask_arr: NDArray[np.bool_] | None = None
    """Mask of pixels to clean"""


class RMCleanOptions(NamedTuple):
    """Options for RM-CLEAN"""

    mask: float
    """Masking threshold - pixels below this value are not cleaned"""
    threshold: float
    """Cleaning threshold - stop when all pixels are below this value"""
    max_iter: int = 1000
    """Maximum clean iterations"""
    gain: float = 0.1
    """Clean loop gain"""
    mask_arr: NDArray[np.bool_] | None = None
    """Additional mask of pixels to avoid"""


def rmclean(
    dirty_fdf_arr: NDArray[np.complex128],
    phi_arr_radm2: NDArray[np.float64],
    rmsf_arr: NDArray[np.complex128],
    phi_double_arr_radm2: NDArray[np.float64],
    fwhm_rmsf_arr: NDArray[np.float64],
    mask: float,
    threshold: float,
    max_iter: int = 1000,
    gain: float = 0.1,
    mask_arr: NDArray[np.bool_] | None = None,
) -> RMCleanResults:
    """Perform RM-CLEAN on a Faraday dispersion function array.

    Args:
        dirty_fdf_arr (NDArray[np.complex128]): Dirty Faraday dispersion function array
        phi_arr_radm2 (NDArray[np.float64]): Faraday depth array in rad/m^2
        rmsf_arr (NDArray[np.complex128]): RMSF array
        phi_double_arr_radm2 (NDArray[np.float64]): Double-length Faraday depth array in rad/m^2
        fwhm_rmsf_arr (NDArray[np.float64]): FWHM of the RMSF array
        mask (float): Masking threshold - pixels below this value are not cleaned
        threshold (float): Cleaning threshold - stop when all pixels are below this value
        max_iter (int, optional): Maximum clean iterations. Defaults to 1000.
        gain (float, optional): Glean loop gain. Defaults to 0.1.
        mask_arr (NDArray[np.bool_] | None, optional): Additional mask of pixels to avoid. Defaults to None.

    Returns:
        RMCleanResults: clean_fdf_arr, model_fdf_arr, clean_iter_arr, resid_fdf_arr
    """

    rm_synth_arrays = RMSynthArrays(
        dirty_fdf_arr=dirty_fdf_arr,
        phi_arr_radm2=phi_arr_radm2,
        rmsf_arr=rmsf_arr,
        phi_double_arr_radm2=phi_double_arr_radm2,
        fwhm_rmsf_arr=fwhm_rmsf_arr,
    )
    clean_options = RMCleanOptions(
        mask=mask,
        threshold=threshold,
        max_iter=max_iter,
        gain=gain,
        mask_arr=mask_arr,
    )

    return _rmclean_nd(rm_synth_arrays, clean_options)


def _rmclean_nd(
    rm_synth_arrays: RMSynthArrays,
    clean_options: RMCleanOptions,
) -> RMCleanResults:
    # Sanity checks on array sizes
    checks: list[tuple[bool, str]] = [
        (
            rm_synth_arrays.phi_arr_radm2.shape[0]
            == rm_synth_arrays.dirty_fdf_arr.shape[0],
            f"'phi_arr_radm2' (size {rm_synth_arrays.phi_arr_radm2.shape[0]}) and 'dirty_fdf_arr' (size {rm_synth_arrays.dirty_fdf_arr.shape[0]}) are not the same length.",
        ),
        (
            rm_synth_arrays.phi_double_arr_radm2.shape[0]
            == rm_synth_arrays.rmsf_arr.shape[0],
            f"Mismatch in 'phi_double_arr_radm2' (size {rm_synth_arrays.phi_double_arr_radm2.shape[0]}) and 'rmsf_arr' (size {rm_synth_arrays.rmsf_arr.shape[0]}) length.",
        ),
        (
            len(rm_synth_arrays.phi_double_arr_radm2)
            >= 2 * len(rm_synth_arrays.phi_arr_radm2),
            f"The Faraday depth of the RMSF (size {len(rm_synth_arrays.phi_double_arr_radm2)}) must be at least twice the FDF (size {len(rm_synth_arrays.phi_arr_radm2)}).",
        ),
        (
            rm_synth_arrays.dirty_fdf_arr.ndim <= 3,
            f"FDF array dimensions ({rm_synth_arrays.dirty_fdf_arr.ndim}) must be <= 3.",
        ),
        (
            rm_synth_arrays.dirty_fdf_arr.ndim == rm_synth_arrays.rmsf_arr.ndim,
            f"The input RMSF (ndim {rm_synth_arrays.rmsf_arr.ndim}) and FDF (ndim {rm_synth_arrays.dirty_fdf_arr.ndim}) must have the same number of axes.",
        ),
        (
            rm_synth_arrays.rmsf_arr.shape[1:]
            == rm_synth_arrays.dirty_fdf_arr.shape[1:],
            f"The xy dimensions of the RMSF {rm_synth_arrays.rmsf_arr.shape[1:]} and FDF {rm_synth_arrays.dirty_fdf_arr.shape[1:]} must match.",
        ),
    ]
    if clean_options.mask_arr is not None:
        checks.append(
            (
                clean_options.mask_arr.shape == rm_synth_arrays.dirty_fdf_arr.shape[1:],
                f"Mask array dimensions {clean_options.mask_arr.shape} must match the xy dimensions of the FDF cube {rm_synth_arrays.dirty_fdf_arr.shape[1:]}.",
            )
        )

    for condition, error_msg in checks:
        if not condition:
            raise ValueError(error_msg)

    # Reshape the arrays to 2D i.e. [phi, x, y] -> [phi, x*y]
    dirty_fdf_arr_2d = nd_to_two_d(rm_synth_arrays.dirty_fdf_arr)
    rmsf_arr_2d = nd_to_two_d(rm_synth_arrays.rmsf_arr)
    iter_count_arr_2d = np.zeros(dirty_fdf_arr_2d.shape[1:], dtype=int)
    fwhm_rmsf_arr_2d = nd_to_two_d(rm_synth_arrays.fwhm_rmsf_arr)

    # Initialise arrays to hold the residual FDF, clean components, clean FDF
    # Residual is initially copies of dirty FDF, so that pixels that are not
    #  processed get correct values (but will be overridden when processed)
    clean_fdf_spectrum_2d = np.zeros_like(dirty_fdf_arr_2d)
    model_fdf_spectrum_2d = np.zeros(dirty_fdf_arr_2d.shape, dtype=complex)
    resid_fdf_arr_2d = dirty_fdf_arr_2d.copy()

    # Loop through the pixels containing a polarised signal
    for pix_idx in tqdm(range(dirty_fdf_arr_2d.shape[1])):
        clean_loop_results = minor_cycle(
            rm_synth_1d_arrays=RMSynthArrays(
                dirty_fdf_arr=resid_fdf_arr_2d[:, pix_idx],
                phi_arr_radm2=rm_synth_arrays.phi_arr_radm2,
                rmsf_arr=rmsf_arr_2d[:, pix_idx],
                phi_double_arr_radm2=rm_synth_arrays.phi_double_arr_radm2,
                fwhm_rmsf_arr=fwhm_rmsf_arr_2d,
                fdf_mask_arr=nd_to_two_d(rm_synth_arrays.fdf_mask_arr)
                if rm_synth_arrays.fdf_mask_arr is not None
                else None,
            ),
            clean_options=clean_options,
        )
        clean_fdf_spectrum_2d[:, pix_idx] = clean_loop_results.clean_fdf_spectrum
        resid_fdf_arr_2d[:, pix_idx] = clean_loop_results.resid_fdf_spectrum
        model_fdf_spectrum_2d[:, pix_idx] = clean_loop_results.model_fdf_spectrum
        iter_count_arr_2d[pix_idx] = clean_loop_results.iter_count

    # Restore the residual to the cleaned FDF (moved outside of loop:
    # will now work for pixels/spectra without clean components)
    clean_fdf_spectrum_2d += resid_fdf_arr_2d

    # Reshape the arrays back to their original shape
    clean_fdf_spectrum = two_d_to_nd(
        clean_fdf_spectrum_2d, rm_synth_arrays.dirty_fdf_arr.shape
    )
    model_fdf_spectrum = two_d_to_nd(
        model_fdf_spectrum_2d, rm_synth_arrays.dirty_fdf_arr.shape
    )
    if rm_synth_arrays.dirty_fdf_arr.shape[1:] == ():
        iter_count_arr = two_d_to_nd(iter_count_arr_2d, (1,))
    else:
        iter_count_arr = two_d_to_nd(
            iter_count_arr_2d, rm_synth_arrays.dirty_fdf_arr.shape[1:]
        )
    resid_fdf_arr = two_d_to_nd(resid_fdf_arr_2d, rm_synth_arrays.dirty_fdf_arr.shape)

    return RMCleanResults(
        clean_fdf_spectrum, model_fdf_spectrum, iter_count_arr, resid_fdf_arr
    )


class MinorLoopArrays(NamedTuple):
    """Arrays for the RM-CLEAN minor loop"""

    resid_fdf_spectrum_mask: np.ma.MaskedArray
    """Residual Faraday dispersion function spectrum"""
    phi_arr_radm2: NDArray[np.float64]
    """Faraday depth array in rad/m^2"""
    phi_double_arr_radm2: NDArray[np.float64]
    """Double-length Faraday depth array in rad/m^2"""
    rmsf_spectrum: NDArray[np.complex128]
    """RMSF spectrum"""
    rmsf_fwhm: float
    """FWHM of the RMSF"""
    peak_find_arr: NDArray[np.float64] | None = None
    """Peak finding array"""

    def with_options(self, **kwargs):
        as_dict = self._asdict()
        as_dict.update(kwargs)

        return MinorLoopArrays(**as_dict)


class MinorLoopOptions(NamedTuple):
    """Options for the RM-CLEAN minor loop"""

    max_iter: int
    """Maximum number of iterations"""
    gain: float
    """Loop gain"""
    mask: float
    """Masking threshold"""
    threshold: float
    """Threshold for stopping the loop"""
    start_iter: int = 0
    """Starting iteration"""
    update_mask: bool = True
    """Update the mask after each iteration"""

    def with_options(self, **kwargs):
        as_dict = self._asdict()
        as_dict.update(kwargs)

        return MinorLoopOptions(**as_dict)


def minor_loop(
    minor_loop_arrays: MinorLoopArrays,
    minor_loop_options: MinorLoopOptions,
) -> MinorLoopResults:
    # Trust nothing
    resid_fdf_spectrum_mask = minor_loop_arrays.resid_fdf_spectrum_mask.copy()
    resid_fdf_spectrum = minor_loop_arrays.resid_fdf_spectrum_mask.data.copy()
    model_fdf_spectrum = np.zeros_like(resid_fdf_spectrum)
    clean_fdf_spectrum = np.zeros_like(resid_fdf_spectrum)
    model_rmsf_spectrum = np.zeros_like(resid_fdf_spectrum)
    rmsf_spectrum = minor_loop_arrays.rmsf_spectrum.copy()
    phi_arr_radm2 = minor_loop_arrays.phi_arr_radm2.copy()
    mask_arr = ~resid_fdf_spectrum_mask.mask.copy()
    mask_arr_original = mask_arr.copy()
    iter_count = int(minor_loop_options.start_iter)

    if minor_loop_arrays.peak_find_arr is not None:
        peak_find_arr = minor_loop_arrays.peak_find_arr.copy()
        peak_find_arr_mask = np.ma.array(
            minor_loop_arrays.peak_find_arr.copy(), mask=~mask_arr
        )
    else:
        peak_find_arr_mask = None

    # Find the index of the peak of the RMSF
    max_rmsf_index = np.nanargmax(np.abs(rmsf_spectrum))

    # Calculate the padding in the sampled RMSF
    # Assumes only integer shifts and symmetric RMSF
    n_phi_pad = int(
        (len(minor_loop_arrays.phi_double_arr_radm2) - len(phi_arr_radm2)) / 2
    )

    logger.info(f"Starting minor loop... {mask_arr.sum()} pixels in the mask")
    for iter_count in range(
        minor_loop_options.start_iter, minor_loop_options.max_iter + 1
    ):
        if resid_fdf_spectrum_mask.mask.all():
            logger.warning(
                f"All channels masked. Exiting loop...performed {iter_count} iterations"
            )
            break
        if iter_count == minor_loop_options.max_iter:
            logger.warning(
                f"Max iterations reached. Exiting loop...performed {iter_count} iterations"
            )
            break
        if np.ma.max(np.ma.abs(resid_fdf_spectrum_mask)) < minor_loop_options.threshold:
            logger.info(
                f"Threshold reached. Exiting loop...performed {iter_count} iterations"
            )
            break
        # Get the absolute peak channel, values and Faraday depth
        if peak_find_arr_mask is not None:
            peak_fdf_index = np.ma.argmax(np.abs(peak_find_arr_mask))
        else:
            peak_fdf_index = np.ma.argmax(np.abs(resid_fdf_spectrum_mask))
        peak_fdf = resid_fdf_spectrum_mask[peak_fdf_index]
        peak_rm = phi_arr_radm2[peak_fdf_index]

        # A clean component is "loop-gain * peak_fdf
        clean_component = minor_loop_options.gain * peak_fdf
        model_fdf_spectrum[peak_fdf_index] += clean_component

        # At which channel is the clean_component located at in the RMSF?
        peak_rmsf_index = peak_fdf_index + n_phi_pad

        # Shift the RMSF & clip so that its peak is centred above this clean_component
        shifted_rmsf_spectrum = np.roll(
            rmsf_spectrum, peak_rmsf_index - max_rmsf_index
        )[n_phi_pad:-n_phi_pad]
        model_rmsf_spectrum += clean_component * shifted_rmsf_spectrum

        # Subtract the product of the clean_component shifted RMSF from the residual FDF
        resid_fdf_spectrum -= clean_component * shifted_rmsf_spectrum
        if minor_loop_arrays.peak_find_arr is not None:
            peak_find_arr -= np.abs(clean_component * shifted_rmsf_spectrum)

        # Restore the clean_component * a Gaussian to the cleaned FDF
        clean_fdf_spectrum += gaussian(
            x=phi_arr_radm2,
            amplitude=clean_component,
            mean=float(peak_rm),
            fwhm=minor_loop_arrays.rmsf_fwhm,
        )
        # Remake masked residual FDF
        if minor_loop_options.update_mask:
            mask_arr = np.abs(resid_fdf_spectrum) > minor_loop_options.mask
            # Mask anything that was previously masked
            mask_arr = mask_arr & mask_arr_original
        resid_fdf_spectrum_mask = np.ma.array(resid_fdf_spectrum, mask=~mask_arr)
        if peak_find_arr_mask is not None:
            peak_find_arr_mask = np.ma.array(peak_find_arr, mask=~mask_arr)

    return MinorLoopResults(
        clean_fdf_spectrum=clean_fdf_spectrum,
        resid_fdf_spectrum=resid_fdf_spectrum,
        resid_fdf_spectrum_mask=resid_fdf_spectrum_mask,
        model_fdf_spectrum=model_fdf_spectrum,
        model_rmsf_spectrum=model_rmsf_spectrum,
        iter_count=iter_count,
    )


def minor_cycle(
    rm_synth_1d_arrays: RMSynthArrays,
    clean_options: RMCleanOptions,
) -> CleanLoopResults:
    for array in (
        rm_synth_1d_arrays.dirty_fdf_arr,
        rm_synth_1d_arrays.phi_arr_radm2,
        rm_synth_1d_arrays.phi_double_arr_radm2,
    ):
        if array.ndim != 1:
            msg = "Arrays in minor cycle must be 1D."
            raise ValueError(msg)

    # Initialise arrays to hold the residual FDF, clean components, clean FDF
    resid_fdf_spectrum = rm_synth_1d_arrays.dirty_fdf_arr.copy()

    mask_arr = np.abs(rm_synth_1d_arrays.dirty_fdf_arr) > clean_options.mask

    if rm_synth_1d_arrays.fdf_mask_arr is not None:
        assert rm_synth_1d_arrays.fdf_mask_arr.ndim == 1, (
            "Arrays in minor cycle must be 1D."
        )
        mask_arr = np.logical_and(
            mask_arr,
            rm_synth_1d_arrays.fdf_mask_arr.astype(bool),
        )

    resid_fdf_spectrum_mask = np.ma.array(resid_fdf_spectrum, mask=~mask_arr)

    minor_loop_arrays = MinorLoopArrays(
        resid_fdf_spectrum_mask=resid_fdf_spectrum_mask,
        phi_arr_radm2=rm_synth_1d_arrays.phi_arr_radm2,
        phi_double_arr_radm2=rm_synth_1d_arrays.phi_double_arr_radm2,
        rmsf_spectrum=rm_synth_1d_arrays.rmsf_arr,
        rmsf_fwhm=float(rm_synth_1d_arrays.fwhm_rmsf_arr.squeeze()),
    )

    minor_loop_options = MinorLoopOptions(
        max_iter=clean_options.max_iter,
        gain=clean_options.gain,
        mask=clean_options.mask,
        threshold=clean_options.threshold,
        start_iter=0,
        update_mask=True,
    )

    logger.info("Starting initial minor loop...")
    initial_loop_results = minor_loop(
        minor_loop_arrays=minor_loop_arrays,
        minor_loop_options=minor_loop_options,
    )

    # Deep clean
    # Mask where clean components have been added
    mask_arr = np.abs(initial_loop_results.model_fdf_spectrum) > 0
    resid_fdf_spectrum = initial_loop_results.resid_fdf_spectrum
    resid_fdf_spectrum_mask = np.ma.array(resid_fdf_spectrum, mask=~mask_arr)

    logger.info("Initial loop complete. Starting deep clean...")

    deep_loop_results = minor_loop(
        minor_loop_arrays=minor_loop_arrays.with_options(
            resid_fdf_spectrum_mask=resid_fdf_spectrum_mask
        ),
        minor_loop_options=minor_loop_options.with_options(
            start_iter=initial_loop_results.iter_count,
            update_mask=False,
        ),
    )

    clean_fdf_spectrum = np.squeeze(
        deep_loop_results.clean_fdf_spectrum + initial_loop_results.clean_fdf_spectrum
    )
    resid_fdf_spectrum = np.squeeze(deep_loop_results.resid_fdf_spectrum)
    model_fdf_spectrum = np.squeeze(
        deep_loop_results.model_fdf_spectrum + initial_loop_results.model_fdf_spectrum
    )

    return CleanLoopResults(
        clean_fdf_spectrum=clean_fdf_spectrum,
        resid_fdf_spectrum=resid_fdf_spectrum,
        model_fdf_spectrum=model_fdf_spectrum,
        iter_count=deep_loop_results.iter_count,
    )



# #!/usr/bin/env python

# # =============================================================================#
# #                                                                             #
# # NAME:     do_QUfit_1D_nest.py                                               #
# #                                                                             #
# # PURPOSE:  Code to simultaneously fit Stokes Q/I and U/I spectra with a      #
# #           Faraday active models.                                            #
# #                                                                             #
# # MODIFIED: 12-Mar-2018 by C. Purcell                                         #
# #                                                                             #
# # CONTENTS:                                                                   #
# #                                                                             #
# #   main           ... parse command line arguments and initiate procedure    #
# #   run_qufit      ... main function of the QU-fitting procedure              #
# #   prior_call     ... return the prior transform function given prior limits  #
# #   lnlike_call    ... return a function to evaluate ln(like) given the data  #
# #   chisq_model    ... calculate the chi-squared for the model                #
# #   wrap_chains    ... wrap and shift chains of periodic parameters           #
# #   init_mnest     ... initialise multinest using a default dictionary        #
# #   merge_two_dicts .. merge two dictionaries                                 #
# #                                                                             #
# # =============================================================================#
# #                                                                             #
# # The MIT License (MIT)                                                       #
# #                                                                             #
# # Copyright (c) 2018 Cormac R. Purcell                                        #
# #                                                                             #
# # Permission is hereby granted, free of charge, to any person obtaining a     #
# # copy of this software and associated documentation files (the "Software"),  #
# # to deal in the Software without restriction, including without limitation   #
# # the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
# # and/or sell copies of the Software, and to permit persons to whom the       #
# # Software is furnished to do so, subject to the following conditions:        #
# #                                                                             #
# # The above copyright notice and this permission notice shall be included in  #
# # all copies or substantial portions of the Software.                         #
# #                                                                             #
# # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
# # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
# # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE #
# # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      #
# # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
# # FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
# # DEALINGS IN THE SOFTWARE.                                                   #
# #                                                                             #
# # =============================================================================#

# import argparse
# import importlib
# import json
# import os
# import sys
# import time
# import traceback

# import bilby
# import matplotlib.pyplot as plt
# import numpy as np
# from utils.util_plotTk import CustomNavbar, plot_Ipqu_spectra_fig

# from rm_lite.tools_1d.rmsynth import readFile
# from rm_lite.utils.fitting import calculate_StokesI_model, create_frac_spectra, toscalar

# C = 2.997924538e8  # Speed of light [m/s]


# # -----------------------------------------------------------------------------#
# def main():
#     """
#     Start the run_qufit procedure if called from the command line.

#     Run QU-fitting on polarised spectra (1D) stored in an ASCII file. The
#     Stokes I spectra is first fit with a polynomial and the resulting model
#     used to create fractional q = Q/I and u = U/I spectra. If the 'noStokesI'
#     option is given, the input data are assumed to be fractional already.

#     The script uses the Nested Sampling algorithm (Skilling 2004) to find
#     the best fitting parameters, given a prior function on each free parameter.
#     The sampling algorithm also calculates the Bayesian evidence, which can
#     be used for model comparison. Factors of >10 between models mean that one
#     model is strongly favoured over the other.

#     Models and priors are  specified as Python code in files called 'mX.py'
#     within the 'models_ns' directory. See the existing files for examples
#     drawn from the paper Sokoloff et al. 1998, MNRAS 229, pg 189.

#     Main algorithm handles the command line interface, passing all arguments
#     one to run_qufit().
#     """

#     # Help string to be shown using the -h option
#     descStr = """
#     Run QU-fitting on polarised spectra (1D) stored in an ASCII file. The
#     Stokes I spectra is first fit with a polynomial and the resulting model
#     used to create fractional q = Q/I and u = U/I spectra. If the 'noStokesI'
#     option is given, the input data are assumed to be fractional already.

#     The script uses the Nested Sampling algorithm (Skilling 2004) to find
#     the best fitting parameters, given a prior function on each free parameter.
#     The sampling algorithm also calculates the Bayesian evidence, which can
#     be used for model comparison. Factors of >10 between models mean that one
#     model is strongly favoured over the other.

#     Models and priors are  specified as Python code in files called 'mX.py'
#     within the 'models_ns' directory. See the existing files for examples
#     drawn from the paper Sokoloff et al. 1998, MNRAS 229, pg 189.
#     """

#     # Parse the command line options
#     parser = argparse.ArgumentParser(
#         description=descStr, formatter_class=argparse.RawTextHelpFormatter
#     )
#     parser.add_argument(
#         "dataFile",
#         metavar="dataFile.dat",
#         nargs=1,
#         help="ASCII file containing Stokes spectra & errors.",
#     )
#     parser.add_argument(
#         "-m", dest="modelNum", type=int, default=1, help="model number to fit [1]."
#     )
#     parser.add_argument(
#         "-f",
#         dest="fit_function",
#         type=str,
#         default="log",
#         help="Stokes I fitting function: 'linear' or ['log'] polynomials.",
#     )
#     parser.add_argument(
#         "-o",
#         dest="polyOrd",
#         type=int,
#         default=2,
#         help="polynomial order to fit to I spectrum: 0-5 supported, 2 is default.\nSet to negative number to enable dynamic order selection.",
#     )
#     parser.add_argument(
#         "-i",
#         dest="noStokesI",
#         action="store_true",
#         help="ignore the Stokes I spectrum [False].",
#     )
#     parser.add_argument(
#         "-p", dest="showPlots", action="store_true", help="show the plots [False]."
#     )
#     parser.add_argument(
#         "-d",
#         dest="debug",
#         action="store_true",
#         help="turn on debugging messages/plots [False].",
#     )
#     parser.add_argument(
#         "-v", dest="verbose", action="store_true", help="verbose mode [False]."
#     )
#     parser.add_argument(
#         "--sampler",
#         type=str,
#         default="dynesty",
#         help="Which sampler to use with Bilby (see https://lscsoft.docs.ligo.org/bilby/samplers.html) [dynesty].",
#     )
#     parser.add_argument(
#         "--ncores",
#         type=int,
#         default=1,
#         help="Number of cores to use for sampling [1].",
#     )
#     parser.add_argument(
#         "--nlive",
#         type=int,
#         default=300,
#         help="Number of live points to use for sampling [300].",
#     )
#     args = parser.parse_args()

#     # Sanity checks
#     if not os.path.exists(args.dataFile[0]):
#         print("File does not exist: '%s'." % args.dataFile[0])
#         sys.exit()
#     (prefixOut,) = os.path.splitext(args.dataFile[0])

#     data = readFile(args.dataFile[0], 32, verbose=args.verbose, debug=args.debug)

#     # Run the QU-fitting procedure
#     run_qufit(
#         data=data,
#         modelNum=args.modelNum,
#         polyOrd=args.polyOrd,
#         nBits=32,
#         noStokesI=args.noStokesI,
#         showPlots=args.showPlots,
#         debug=args.debug,
#         verbose=args.verbose,
#         fit_function=args.fit_function,
#         sampler=args.sampler,
#         ncores=args.ncores,
#         nlive=args.nlive,
#         prefixOut=prefixOut,
#     )


# def load_model(modelNum, verbose=False):
#     if verbose:
#         print("\nLoading the model from 'models_ns/m%d.py' ..." % modelNum)
#     # First check the working directory for a model. Failing that, try the install directory.
#     try:
#         spec = importlib.util.spec_from_file_location(
#             "m%d" % modelNum, "models_ns/m%d.py" % modelNum
#         )
#         mod = importlib.util.module_from_spec(spec)
#         sys.modules[mod] = mod
#         spec.loader.exec_module(mod)
#     except FileNotFoundError:
#         try:
#             RMtools_dir = os.path.dirname(importlib.util.find_spec("RMtools_1D").origin)
#             spec = importlib.util.spec_from_file_location(
#                 "m%d" % modelNum, RMtools_dir + "/models_ns/m%d.py" % modelNum
#             )
#             mod = importlib.util.module_from_spec(spec)
#             sys.modules[mod] = mod
#             spec.loader.exec_module(mod)
#         except:
#             print(
#                 "Model could not be found! Please make sure model is present either in {}/models_ns/, or in {}/models_ns/".format(
#                     os.getcwd(), RMtools_dir
#                 )
#             )
#             sys.exit()
#     return mod


# # -----------------------------------------------------------------------------#
# def run_qufit(
#     data,
#     modelNum,
#     polyOrd=2,
#     nBits=32,
#     noStokesI=False,
#     showPlots=False,
#     debug=False,
#     verbose=False,
#     sampler="dynesty",
#     fit_function="log",
#     ncores=1,
#     nlive=1000,
#     prefixOut="prefixOut",
# ):
#     """Carry out QU-fitting using the supplied parameters:
#     data (list): Contains frequency and polarization data as either:
#         [freq_Hz, I, Q, U, dI, dQ, dU]
#             freq_Hz (array_like): Frequency of each channel in Hz.
#             I (array_like): Stokes I intensity in each channel.
#             Q (array_like): Stokes Q intensity in each channel.
#             U (array_like): Stokes U intensity in each channel.
#             dI (array_like): Error in Stokes I intensity in each channel.
#             dQ (array_like): Error in Stokes Q intensity in each channel.
#             dU (array_like): Error in Stokes U intensity in each channel.
#         or
#         [freq_Hz, q, u,  dq, du]
#             freq_Hz (array_like): Frequency of each channel in Hz.
#             q (array_like): Fractional Stokes Q intensity (Q/I) in each channel.
#             u (array_like): Fractional Stokes U intensity (U/I) in each channel.
#             dq (array_like): Error in fractional Stokes Q intensity in each channel.
#             du (array_like): Error in fractional Stokes U intensity in each channel.
#     modelNum (int, required): number of model to be fit to data. Models and
#          priors are specified as Python code in files called 'mX.py' within
#         the 'models_ns' directory.
#     outDir (str): relative or absolute path to save outputs to. Defaults to
#         working directory.
#     polyOrd (int): Order of polynomial to fit to Stokes I spectrum (used to
#         normalize Q and U values). Defaults to 3 (cubic).
#     nBits (int): number of bits to use in internal calculations.
#     noStokesI (bool): set True if the Stokes I spectrum should be ignored.
#     showPlots (bool): Set true if the spectrum and parameter space plots
#         should be displayed.
#     sigma_clip (float): How many standard deviations to clip around the
#         mean of each mode in the parameter postierors.
#     debug (bool): Display debug messages.
#     verbose (bool): Print verbose messages/results to terminal.

#     Returns: nothing. Results saved to files and/or printed to terminal."""

#     # Output prefix is derived from the input file name
#     nestOut = f"{prefixOut}_m{modelNum}_{sampler}/"

#     # Parse the data array
#     # freq_Hz, I, Q, U, dI, dQ, dU
#     try:
#         (freqArr_Hz, IArr, QArr, UArr, dIArr, dQArr, dUArr) = data
#         print("\nFormat [freq_Hz, I, Q, U, dI, dQ, dU]")
#     except Exception:
#         # freq_Hz, Q, U, dQ, dU
#         try:
#             (freqArr_Hz, QArr, UArr, dQArr, dUArr) = data
#             print("\nFormat [freq_Hz, Q, U,  dQ, dU]")
#             noStokesI = True
#         except Exception:
#             print("\nError: Failed to parse data file!")
#             if debug:
#                 print(traceback.format_exc())
#             return

#     # If no Stokes I present, create a dummy spectrum = unity
#     if noStokesI:
#         print("Note: no Stokes I data - assuming fractional polarisation.")
#         IArr = np.ones_like(QArr)
#         dIArr = np.zeros_like(QArr)

#     # Convert to GHz for convenience
#     lamSqArr_m2 = np.power(C / freqArr_Hz, 2.0)

#     # Fit the Stokes I spectrum and create the fractional spectra
#     dataArr = create_frac_spectra(
#         freqArr=freqArr_Hz,
#         IArr=IArr,
#         QArr=QArr,
#         UArr=UArr,
#         dIArr=dIArr,
#         dQArr=dQArr,
#         dUArr=dUArr,
#         polyOrd=polyOrd,
#         verbose=True,
#         fit_function=fit_function,
#     )
#     (IModArr, qArr, uArr, dqArr, duArr, fit_result) = dataArr

#     # Plot the data and the Stokes I model fit
#     print("Plotting the input data and spectral index fit.")
#     freqHirArr_Hz = np.linspace(freqArr_Hz[0], freqArr_Hz[-1], 10000)
#     IModHirArr = calculate_StokesI_model(fit_result, freqHirArr_Hz)
#     specFig = plt.figure(facecolor="w", figsize=(10, 6))
#     plot_Ipqu_spectra_fig(
#         freqArr_Hz=freqArr_Hz,
#         IArr=IArr,
#         qArr=qArr,
#         uArr=uArr,
#         dIArr=dIArr,
#         dqArr=dqArr,
#         duArr=duArr,
#         freqHirArr_Hz=freqHirArr_Hz,
#         IModArr=IModHirArr,
#         fig=specFig,
#     )

#     # Use the custom navigation toolbar
#     try:
#         specFig.canvas.toolbar.pack_forget()
#         CustomNavbar(specFig.canvas, specFig.canvas.toolbar.window)
#     except Exception:
#         pass

#     # Display the figure
#     if showPlots:
#         specFig.canvas.draw()
#         specFig.show()

#     # -------------------------------------------------------------------------#

#     # Load the model and parameters from the relevant file
#     mod = load_model(modelNum, verbose=True)

#     model = mod.model

#     # Let's time the sampler
#     startTime = time.time()

#     parNames = []
#     priorTypes = []
#     labels = []
#     bounds = []
#     wraps = []
#     for key, prior in mod.priors.items():
#         if prior.__class__.__name__ == "Constraint":
#             continue
#         parNames.append(key)
#         priorTypes.append(prior.__class__.__name__)
#         labels.append(prior.latex_label)
#         bounds.append([prior.minimum, prior.maximum])
#         wraps.append(prior.boundary)
#     nDim = len(parNames)
#     fixedMsk = [0 if x == "DeltaFunction" else 1 for x in priorTypes]
#     nFree = sum(fixedMsk)

#     # Set the prior function given the bounds of each parameter
#     priors = mod.priors

#     # Set the likelihood function given the data
#     lnlike = lnlike_call(parNames, lamSqArr_m2, qArr, dqArr, uArr, duArr, modelNum)
#     # Let's time the sampler
#     startTime = time.time()

#     result = bilby.run_sampler(
#         likelihood=lnlike,
#         priors=priors,
#         sampler=sampler,
#         nlive=nlive,
#         npool=ncores,
#         outdir=nestOut,
#         label="m%d" % modelNum,
#         plot=True,
#     )

#     # Do the post-processing on one processor
#     endTime = time.time()

#     # Shift the angles by 90 deg, if necessary
#     # This is to get around the angle wrap issue
#     rotated_parNames = []  ## Store parameters that have been rotated
#     for i in range(nDim):
#         if (
#             parNames[i][-3:] == "deg"
#             and bounds[i][0] == 0.0
#             and bounds[i][1] == 180.0
#             and wraps[i] == "periodic"
#         ):
#             # Only try to do it if the units is in deg, bounded within [0., 180.], and is a periodic variable
#             summary_tmp = result.get_one_dimensional_median_and_error_bar(parNames[i])
#             if summary_tmp.median < 45.0 or summary_tmp.median >= 135.0:
#                 # Shift PA by 90 deg
#                 result.samples[:, i] += 90.0
#                 # Keep all values within [0, 180)
#                 result.samples[:, i] -= (result.samples[:, i] >= 180.0) * 180.0
#                 # Keep track of which parameters have been rotated
#                 rotated_parNames.append(parNames[i])

#     # Update the posterior values
#     if len(rotated_parNames) > 0:
#         result.samples_to_posterior()

#     # Best guess here - taking the maximum likelihood value
#     lnLike = np.max(result.log_likelihood_evaluations)
#     lnEvidence = result.log_evidence
#     dLnEvidence = result.log_evidence_err

#     # Get the best-fitting values & uncertainties

#     p = [None] * nDim
#     errPlus = [None] * nDim
#     errMinus = [None] * nDim
#     # g = lambda v: (v[1], v[2]-v[1], v[1]-v[0])
#     for i in range(nDim):
#         summary = result.get_one_dimensional_median_and_error_bar(parNames[i])
#         # Get stats around modal value
#         # Shift back by 90 deg if necessary
#         median_val = summary.median - 90.0 * (parNames[i] in rotated_parNames)
#         median_val += 180.0 * (median_val < 0.0) * (parNames[i] in rotated_parNames)
#         plus_val = summary.plus
#         minus_val = summary.minus
#         p[i], errPlus[i], errMinus[i] = (
#             median_val,
#             plus_val,
#             minus_val,
#         )

#     # Calculate goodness-of-fit parameters
#     nData = 2.0 * len(lamSqArr_m2)
#     dof = nData - nFree - 1
#     chiSq = chisq_model(parNames, p, lamSqArr_m2, qArr, dqArr, uArr, duArr, model)
#     chiSqRed = chiSq / dof
#     AIC = 2.0 * nFree - 2.0 * lnLike
#     AICc = 2.0 * nFree * (nFree + 1) / (nData - nFree - 1) - 2.0 * lnLike
#     BIC = nFree * np.log(nData) - 2.0 * lnLike

#     # Summary of run
#     print("")
#     print("-" * 80)
#     print("SUMMARY OF SAMPLING RUN:")
#     print("#-PROCESSORS  = %d" % ncores)
#     print("RUN-TIME      = %.2f" % (endTime - startTime))
#     print("DOF           = %d" % dof)
#     print("CHISQ:        = %.3g" % chiSq)
#     print("CHISQ RED     = %.3g" % chiSqRed)
#     print("AIC:          = %.3g" % AIC)
#     print("AICc          = %.3g" % AICc)
#     print("BIC           = %.3g" % BIC)
#     print("ln(EVIDENCE)  = %.3g" % lnEvidence)
#     print("dLn(EVIDENCE) = %.3g" % dLnEvidence)
#     print("")
#     print("-" * 80)
#     print("RESULTS:\n")
#     for i in range(len(p)):
#         print("%s = %.4g (+%3g, -%3g)" % (parNames[i], p[i], errPlus[i], errMinus[i]))
#     print("-" * 80)
#     print("")

#     # Create a save dictionary and store final p in values
#     outFile = f"{prefixOut}_m{modelNum}_{sampler}.json"
#     saveDict = {
#         "parNames": toscalar(parNames),
#         "labels": toscalar(labels),
#         "values": toscalar(p),
#         "errPlus": toscalar(errPlus),
#         "errMinus": toscalar(errMinus),
#         "bounds": toscalar(bounds),
#         "priorTypes": toscalar(priorTypes),
#         "wraps": toscalar(wraps),
#         "dof": toscalar(dof),
#         "chiSq": toscalar(chiSq),
#         "chiSqRed": toscalar(chiSqRed),
#         "AIC": toscalar(AIC),
#         "AICc": toscalar(AICc),
#         "BIC": toscalar(BIC),
#         "ln(EVIDENCE) ": toscalar(lnEvidence),
#         "dLn(EVIDENCE)": toscalar(dLnEvidence),
#         "nFree": toscalar(nFree),
#         "Imodel": ",".join([str(x.astype(np.float32)) for x in fit_result.params]),
#         "Imodel_errs": ",".join([str(x.astype(np.float32)) for x in fit_result.perror]),
#         "IfitChiSq": toscalar(fit_result.chiSq),
#         "IfitChiSqRed": toscalar(fit_result.chiSqRed),
#         "IfitPolyOrd": toscalar(fit_result.polyOrd),
#         "Ifitfreq0": toscalar(fit_result.reference_frequency_Hz),
#     }

#     for k, v in saveDict.items():
#         if isinstance(v, np.float_):
#             saveDict[k] = float(v)
#         elif isinstance(v, np.int_):
#             saveDict[k] = int(v)
#         elif isinstance(v, NDArray[np.float64]):
#             saveDict[k] = v.tolist()
#         elif isinstance(v, np.bool_):
#             saveDict[k] = bool(v)
#     json.dump(saveDict, open(outFile, "w"))
#     outFile = f"{prefixOut}_m{modelNum}_{sampler}.dat"
#     FH = open(outFile, "w")
#     for k, v in saveDict.items():
#         FH.write("%s=%s\n" % (k, v))
#     FH.close()
#     print("Results saved in JSON and .dat format to:\n '%s'\n" % outFile)

#     # Plot the posterior samples in a corner plot
#     # chains =  aObj.get_equal_weighted_posterior()
#     # chains = wrap_chains(chains, wraps, bounds, p)[:, :nDim]
#     # iFixed = [i for i, e in enumerate(fixedMsk) if e==0]
#     # chains = np.delete(chains, iFixed, 1)
#     # for i in sorted(iFixed, reverse=True):
#     #     del(labels[i])
#     #     del(p[i])

#     # Tricky to get correct PA on the corner plot because of PA wrap!
#     # Solution: Shift PA back to original, and ignore limit of [0, 180]

#     for i in range(nDim):
#         if parNames[i] in rotated_parNames:
#             # Rotate back by the 90 deg
#             result.samples[:, i] -= 90.0
#             if p[i] > 135.0:
#                 # In case the PA shown would be << 0 deg
#                 result.samples[:, i] += 180.0

#     # Resampling to make sure the results show
#     if len(rotated_parNames) > 0:
#         result.samples_to_posterior()

#     cornerFig = result.plot_corner()

#     # Save the posterior chains to ASCII file

#     # Plot the data and best-fitting model
#     lamSqHirArr_m2 = np.linspace(lamSqArr_m2[0], lamSqArr_m2[-1], 10000)
#     freqHirArr_Hz = C / np.sqrt(lamSqHirArr_m2)
#     IModArr = calculate_StokesI_model(fit_result, freqHirArr_Hz)
#     pDict = {k: v for k, v in zip(parNames, p)}
#     quModArr = model(pDict, lamSqHirArr_m2)
#     model_dict = {
#         "model": model,
#         "parNames": parNames,
#         "posterior": result.posterior,
#     }
#     specFig.clf()
#     plot_Ipqu_spectra_fig(
#         freqArr_Hz=freqArr_Hz,
#         IArr=IArr,
#         qArr=qArr,
#         uArr=uArr,
#         dIArr=dIArr,
#         dqArr=dqArr,
#         duArr=duArr,
#         freqHirArr_Hz=freqHirArr_Hz,
#         IModArr=IModArr,
#         qModArr=quModArr.real,
#         uModArr=quModArr.imag,
#         model_dict=model_dict,
#         fig=specFig,
#     )
#     specFig.canvas.draw()

#     # Save the figures
#     outFile = prefixOut + "fig_m%d_specfit.pdf" % modelNum
#     specFig.set_canvas(specFig.canvas)
#     specFig.savefig(outFile)
#     print("Plot of best-fitting model saved to:\n '%s'\n" % outFile)
#     outFile = prefixOut + "fig_m%d_corner.pdf" % modelNum
#     cornerFig.set_canvas(cornerFig.canvas)
#     cornerFig.savefig(outFile)
#     print("Plot of posterior samples saved to \n '%s'\n" % outFile)

#     # Display the figures
#     if showPlots:
#         specFig.show()
#         cornerFig.show()


# # -----------------------------------------------------------------------------#
# class lnlike_call(bilby.Likelihood):
#     """Returns a function to evaluate the log-likelihood"""

#     def __init__(self, parNames, lamSqArr_m2, qArr, dqArr, uArr, duArr, modelNum):
#         self.parNames = parNames
#         self.lamSqArr_m2 = lamSqArr_m2
#         self.qArr = qArr
#         self.dqArr = dqArr
#         self.uArr = uArr
#         self.duArr = duArr
#         self.modelNum = modelNum
#         pDict = {k: None for k in parNames}
#         super().__init__(parameters=pDict)


# # -----------------------------------------------------------------------------#
# def chisq_model(parNames, p, lamSqArr_m2, qArr, dqArr, uArr, duArr, model):
#     """Calculate the chi^2 for the current model, given the data."""

#     # Evaluate the model and calculate chisq
#     pDict = {k: v for k, v in zip(parNames, p)}
#     quMod = model(pDict, lamSqArr_m2)
#     chisq = np.nansum(
#         np.power((qArr - quMod.real) / dqArr, 2)
#         + np.power((uArr - quMod.imag) / duArr, 2)
#     )

#     return chisq


# # -----------------------------------------------------------------------------#
# # -----------------------------------------------------------------------------#
# # -----------------------------------------------------------------------------#
# # -----------------------------------------------------------------------------#
# if __name__ == "__main__":
#     main()



# #!/usr/bin/env python3
# # =============================================================================#
# #                                                                             #
# # NAME:     rmtools_bwpredict.py                                              #
# #                                                                             #
# # PURPOSE: Algorithm for finding polarized sources while accounting for       #
# #          bandwidth depolarization.                                          #
# #                                                                             #
# # =============================================================================#
# #                                                                             #
# # The MIT License (MIT)                                                       #
# #                                                                             #
# # Copyright (c) 2020 Canadian Initiative for Radio Astronomy Data Analysis    #                                                                             #
# # Permission is hereby granted, free of charge, to any person obtaining a     #
# # copy of this software and associated documentation files (the "Software"),  #
# # to deal in the Software without restriction, including without limitation   #
# # the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
# # and/or sell copies of the Software, and to permit persons to whom the       #
# # Software is furnished to do so, subject to the following conditions:        #
# #                                                                             #
# # The above copyright notice and this permission notice shall be included in  #
# # all copies or substantial portions of the Software.                         #
# #                                                                             #
# # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
# # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
# # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE #
# # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      #
# # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
# # FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
# # DEALINGS IN THE SOFTWARE.                                                   #
# #                                                                             #
# # =============================================================================#

# import argparse
# import math as m
# import sys

# import matplotlib.pyplot as plt
# import numpy as np

# from rm_lite.tools_1d.bwdepol import (
#     adjoint_theory,
#     estimate_channel_bandwidth,
#     plot_adjoint_info,
# )

# if sys.version_info.major == 2:
#     print("RM-tools will no longer run with Python 2! Please use Python 3.")
#     exit()
# C = 2.997924538e8  # Speed of light [m/s]


# # -----------------------------------------------------------------------------#
# def bwdepol_compute_predictions(
#     freqArr_Hz, widths_Hz=None, phiMax_radm2=None, dPhi_radm2=None
# ):
#     """Computes theoretical sensitivity and noise curves for given
#     channelization.

#     Parameters
#     ----------
#     freqArr_Hz:  array, float
#                  array of the centers of the frequency channels

#     Kwargs
#     ------
#     phiMax_radm2: float
#                   Maximum absolute Faraday depth (rad/m^2)

#     dPhi_radm2: float
#                 Faraday depth channel size (rad/m^2)

#     nSamples: float
#               Number of samples across the RMSF

#     Returns
#     -------
#     adjoint_info: list
#                   Faraday depth array, sensitivity array, noise array

#     """
#     # Calculate some wavelength parameters
#     lambdaSqArr_m2 = np.power(C / freqArr_Hz, 2.0)
#     # dFreq_Hz = np.nanmin(np.abs(np.diff(freqArr_Hz)))
#     lambdaSqRange_m2 = np.nanmax(lambdaSqArr_m2) - np.nanmin(lambdaSqArr_m2)
#     dLambdaSqMin_m2 = np.nanmin(np.abs(np.diff(lambdaSqArr_m2)))
#     dLambdaSqMax_m2 = np.nanmax(np.abs(np.diff(lambdaSqArr_m2)))

#     # Set the Faraday depth range
#     fwhmRMSF_radm2 = 2.0 * m.sqrt(3.0) / lambdaSqRange_m2
#     if dPhi_radm2 is None:
#         dPhi_radm2 = fwhmRMSF_radm2
#     if phiMax_radm2 is None:
#         phiMax_radm2 = 2 * m.sqrt(3.0) / dLambdaSqMin_m2

#     # Faraday depth sampling.
#     phiArr_radm2 = np.arange(0, phiMax_radm2 + 1e-6, dPhi_radm2)
#     phiArr_radm2 = phiArr_radm2.astype("float64")

#     print(
#         "Computing out to a Faraday depth of {:g} rad/m^2 in steps of {:g} rad/m^2".format(
#             phiMax_radm2, dPhi_radm2
#         )
#     )

#     # Uniform weights only for prediction purposes
#     weightArr = np.ones(freqArr_Hz.shape, dtype="float64")

#     # Get channel widths if not given by user.
#     K = 1.0 / np.sum(weightArr)
#     if widths_Hz is None:
#         widths_Hz = estimate_channel_bandwidth(freqArr_Hz)

#     adjoint_varbs = [widths_Hz, freqArr_Hz, phiArr_radm2, K, weightArr]
#     adjoint_info = adjoint_theory(adjoint_varbs, weightArr, show_progress=False)
#     phiArr_radm2, adjoint_sens, adjoint_noise = adjoint_info

#     adjoint_info[2] = adjoint_noise / np.max(adjoint_noise)  # Renormalize to unity.

#     return adjoint_info


# def main():
#     """
#     Start the function to generate the figures if called from the command line.
#     """
#     # Help string to be shown using the -h option
#     descStr = """
#     Calculate the theoretical sensitivity and noise curves for the bandwidth-
#     depolarization-corrected RM synthesis method described in Fine et al. (2022).

#     Takes in a ASCII file containing either 1 column (channel frequencies, in Hz)
#     or two columns (channel frequencies and channel bandwidths in Hz, space separated).

#     Generates interactive plots of the two curves. These are intended to guide
#     users in deciding in what RM range traditional RM synthesis is sufficiently
#     accurate, and over what RM range they may want to use the modified method.
#     """

#     # Parse the command line options
#     parser = argparse.ArgumentParser(
#         description=descStr, formatter_class=argparse.RawTextHelpFormatter
#     )
#     parser.add_argument(
#         "dataFile",
#         metavar="dataFile.dat",
#         nargs=1,
#         help="ASCII file containing channel frequencies.",
#     )
#     parser.add_argument(
#         "-l",
#         dest="phiMax_radm2",
#         type=float,
#         default=None,
#         help="absolute max Faraday depth sampled [Auto].",
#     )
#     parser.add_argument(
#         "-d",
#         dest="dPhi_radm2",
#         type=float,
#         default=None,
#         help="width of Faraday depth channel [Auto].\n(overrides -s NSAMPLES flag)",
#     )
#     parser.add_argument(
#         "-s",
#         dest="nSamples",
#         type=float,
#         default=10,
#         help="number of samples across the RMSF lobe [10].",
#     )

#     args = parser.parse_args()

#     # Get data:
#     try:
#         data = np.loadtxt(args.dataFile[0], unpack=True, dtype="float64")
#         if data.ndim == 1:  # If single column file, data is only channel freqs
#             freqArr_Hz = data
#             widths_Hz = None
#         else:  # file has multiple columns
#             freqArr_Hz = data[0]  # assume the first column is channel freqs
#             widths_Hz = data[1]  # Assume widths are 2nd column if present.
#     except:
#         print(
#             "Unable to read file. Please ensure file is readable and contains 1 or 2 columns."
#         )
#         exit

#     adjoint_info = bwdepol_compute_predictions(
#         freqArr_Hz=freqArr_Hz,
#         widths_Hz=widths_Hz,
#         phiMax_radm2=args.phiMax_radm2,
#         dPhi_radm2=args.dPhi_radm2,
#     )

#     # plot adjoint info
#     plot_adjoint_info(adjoint_info, units="arb. units")
#     plt.show()


# # -----------------------------------------------------------------------------#
# if __name__ == "__main__":
#     main()



"""RM-CLEAN on 1D data"""

from __future__ import annotations

from typing import NamedTuple

import numpy as np
import polars as pl
from numpy.typing import NDArray
from scipy import interpolate

from rm_lite.tools_1d.rmsynth import RMSynth1DResults
from rm_lite.utils.clean import rmclean
from rm_lite.utils.logging import logger
from rm_lite.utils.synthesis import (
    TheoreticalNoise,
    get_fdf_parameters,
    lambda2_to_freq,
)

rmclean_arrs_schema = pl.Schema(
    {
        "phi_arr_radm2": pl.Float64,
        "fdf_dirty_complex_arr": pl.Object,
        "fdf_clean_complex_arr": pl.Object,
        "fdf_model_complex_arr": pl.Object,
        "fdf_residual_complex_arr": pl.Object,
    }
)
rmclean_arrs_schema_df = rmclean_arrs_schema.to_frame(eager=True)

rmclean_scalar_schema = pl.Schema(
    {
        "mask": pl.Float64,
        "threshold": pl.Float64,
        "n_iter": pl.Int64,
    }
)
rmclean_scalar_schema_df = rmclean_scalar_schema.to_frame(eager=True)


class RMClean1DResults(NamedTuple):
    """Resulting arrays from RM-synthesis"""

    fdf_parameters: pl.DataFrame
    """ FDF parameters """
    fdf_arrs: pl.DataFrame
    """ RMClean arrays """
    clean_parameters: pl.DataFrame
    """ RMClean parameters """


def run_rmclean_from_synth(
    rm_synth_1d_results: RMSynth1DResults,
    auto_mask: float = 7,
    auto_threshold: float = 1,
    max_iter: int = 10_000,
    gain: float = 0.1,
    mask_arr: NDArray[np.bool_] | None = None,
) -> RMClean1DResults:
    """Run RM-CLEAN on the results of RM-synth.

    Args:
        rm_synth_1d_results (RMSynth1DResults): Results from RM-synth.
        auto_mask (float, optional): Masking threshold in SNR. Defaults to 7.
        auto_threshold (float, optional): Cleaning threshold in SNR. Defaults to 1.
        max_iter (int, optional): Maximum CLEAN iterations. Defaults to 10_000.
        gain (float, optional): CLEAN gain. Defaults to 0.1.
        mask_arr (NDArray[np.bool_] | None, optional): Optional mask array. Defaults to None.

    Returns:
        RMClean1DResults: RM-CLEAN results: `fdf_parameters`, `fdf_arrs`, `clean_parameters`.
    """
    rmsyth_arrs_df = rm_synth_1d_results.fdf_arrs
    rmsf_arrs_df = rm_synth_1d_results.rmsf_arrs
    fdf_parameters = rm_synth_1d_results.fdf_parameters
    stokes_i_arrs_df = rm_synth_1d_results.stokes_i_arrs

    theoretical_noise = TheoreticalNoise(
        fdf_error_noise=float(fdf_parameters["fdf_error_noise"].to_numpy().squeeze()),
        fdf_q_noise=float(fdf_parameters["fdf_q_noise"].to_numpy().squeeze()),
        fdf_u_noise=float(fdf_parameters["fdf_u_noise"].to_numpy().squeeze()),
    )

    logger.info(f"Theoretical noise: {theoretical_noise}")

    mask = auto_mask * theoretical_noise.fdf_error_noise
    threshold = auto_threshold * theoretical_noise.fdf_error_noise

    logger.info(
        f"Auto mask: {mask:0.2f}, Auto threshold: {threshold:0.2f}, Max iterations: {max_iter}, Gain: {gain}"
    )

    stokes_i_model = interpolate.interp1d(
        stokes_i_arrs_df["freq_arr_hz"],
        stokes_i_arrs_df["stokes_i_model_arr"],
    )

    stokes_i_reference_flux = stokes_i_model(
        lambda2_to_freq(float(fdf_parameters["lam_sq_0_m2"].to_numpy()[0]))
    )

    fdf_dirty_arr = rmsyth_arrs_df["fdf_dirty_complex_arr"].to_numpy().astype(complex)

    rm_clean_results = rmclean(
        dirty_fdf_arr=fdf_dirty_arr,
        phi_arr_radm2=rmsyth_arrs_df["phi_arr_radm2"].to_numpy().astype(float),
        rmsf_arr=rmsf_arrs_df["rmsf_complex_arr"].to_numpy().astype(complex),
        phi_double_arr_radm2=rmsf_arrs_df["phi2_arr_radm2"].to_numpy().astype(float),
        fwhm_rmsf_arr=fdf_parameters["fwhm_rmsf_radm2"].to_numpy().astype(float),
        mask=mask,
        threshold=threshold,
        max_iter=max_iter,
        gain=gain,
        mask_arr=mask_arr,
    )
    clean_fdf_arr, model_fdf_arr, clean_iter_arr, resid_fdf_arr = rm_clean_results

    fdf_parameters = get_fdf_parameters(
        fdf_arr=rm_clean_results.clean_fdf_arr,
        phi_arr_radm2=rmsyth_arrs_df["phi_arr_radm2"].to_numpy().astype(float),
        fwhm_rmsf_radm2=float(
            fdf_parameters["fwhm_rmsf_radm2"].to_numpy().astype(float).squeeze()
        ),
        freq_arr_hz=stokes_i_arrs_df["freq_arr_hz"].to_numpy().astype(float),
        complex_pol_arr=stokes_i_arrs_df["complex_frac_pol_arr"]
        .to_numpy()
        .astype(complex),
        complex_pol_error=stokes_i_arrs_df["complex_frac_pol_error"]
        .to_numpy()
        .astype(complex),
        lambda_sq_arr_m2=stokes_i_arrs_df["lambda_sq_arr_m2"].to_numpy().astype(float),
        lam_sq_0_m2=float(fdf_parameters["lam_sq_0_m2"].to_numpy().squeeze()),
        stokes_i_reference_flux=stokes_i_reference_flux,
        theoretical_noise=theoretical_noise,
        fit_function=str(fdf_parameters["fit_function"].to_numpy().squeeze()),
    )

    rmclean_arrs = rmclean_arrs_schema_df.vstack(
        pl.DataFrame(
            {
                "phi_arr_radm2": rmsyth_arrs_df["phi_arr_radm2"]
                .to_numpy()
                .astype(float),
                "fdf_dirty_complex_arr": rmsyth_arrs_df["fdf_dirty_complex_arr"]
                .to_numpy()
                .astype(complex),
                "fdf_clean_complex_arr": clean_fdf_arr,
                "fdf_model_complex_arr": model_fdf_arr,
                "fdf_residual_complex_arr": resid_fdf_arr,
            }
        )
    )

    clean_parameters = rmclean_scalar_schema_df.vstack(
        pl.DataFrame(
            {
                "mask": mask,
                "threshold": threshold,
                "n_iter": clean_iter_arr,
            }
        )
    )

    return RMClean1DResults(
        fdf_parameters=fdf_parameters,
        fdf_arrs=rmclean_arrs,
        clean_parameters=clean_parameters,
    )



# #!/usr/bin/env python3
# """
# QA testing tools for RM-tools.
# These tools are intended to produce test simulated data sets, and run them
# through RM-tools. Automated tools will only be able to confirm that things ran,
# but user inspection of the results will be needed to confirm that the expected
# values are produced.

# Random values are necessary to simulate noise, which is expected for different
# parts of the code. I have forced the random seed to be the same each run,
# in order to make the tests deterministic. This works for everything except
# QU fitting, which uses random numbers internally that can't be controlled.

# Created on Fri Oct 25 10:00:24 2019
# @author: Cameron Van Eck
# """
# import json
# import os
# import shutil
# import subprocess
# import unittest
# from pathlib import Path

# import numpy as np
# from astropy.io import fits as pf
# from scipy.ndimage import gaussian_filter

# TEST_PATH = Path(__file__).parent.absolute()
# ONED_PATH = TEST_PATH / "simdata" / "1D"
# THREED_PATH = TEST_PATH / "simdata" / "3D"


# def Faraday_thin_complex_polarization(freq_arr, RM, Polint, initial_angle):
#     """freq_arr = channel frequencies in Hz
#     RM = source RM in rad m^-2
#     Polint = polarized intensity in whatever units
#     initial angle = pre-rotation polarization angle (in degrees)"""
#     l2_arr = (299792458.0 / freq_arr) ** 2
#     Q = Polint * np.cos(2 * (np.outer(l2_arr, RM) + np.deg2rad(initial_angle)))
#     U = Polint * np.sin(2 * (np.outer(l2_arr, RM) + np.deg2rad(initial_angle)))
#     return np.squeeze(np.transpose(Q + 1j * U))


# def create_1D_data(freq_arr):
#     RM = 200
#     pol_angle_deg = 50
#     StokesI_midband = 1
#     fracpol = 0.7
#     noise_amplitude = 0.1
#     spectral_index = -0.7
#     error_estimate = 1  # Size of assumed errors as multiple of actual error.
#     if not ONED_PATH.exists():
#         ONED_PATH.mkdir(parents=True)
#     shutil.copy(TEST_PATH / "RMsynth1D_testdata.dat", ONED_PATH / "simsource.dat")
#     with open(ONED_PATH / "sim_truth.txt", "w") as f:
#         f.write("RM = {} rad/m^2\n".format(RM))
#         f.write("Intrsinsic polarization angle = {} deg\n".format(pol_angle_deg))
#         f.write("Fractional polarization = {} %\n".format(fracpol * 100.0))
#         f.write("Stokes I = {} Jy/beam\n".format(StokesI_midband))
#         f.write(
#             "Reference frequency for I = {} GHz\n".format(np.median(freq_arr) / 1e9)
#         )
#         f.write("Spectral index = {}\n".format(spectral_index))
#         f.write("Actual error per channel = {} Jy/beam\n".format(noise_amplitude))
#         f.write(
#             "Input assumed error = {} Jy/beam\n".format(
#                 noise_amplitude * error_estimate
#             )
#         )


# def create_3D_data(freq_arr, N_side=100):
#     src_RM = 200
#     src_pol_angle_deg = 50
#     src_flux = 2
#     src_x = N_side // 4
#     src_y = N_side // 4

#     diffuse_RM = 50
#     diffuse_pol_angle_deg = -10
#     diffuse_flux = 1

#     noise_amplitude = 0.1
#     beam_size_pix = 20

#     src_pol_spectrum = Faraday_thin_complex_polarization(
#         freq_arr, src_RM, src_flux, src_pol_angle_deg
#     )
#     diffuse_pol_spectrum = Faraday_thin_complex_polarization(
#         freq_arr, diffuse_RM, diffuse_flux, diffuse_pol_angle_deg
#     )

#     src_Q_cube = np.zeros((N_side, N_side, freq_arr.size))
#     src_U_cube = np.zeros((N_side, N_side, freq_arr.size))
#     src_Q_cube[src_x, src_y, :] = src_pol_spectrum.real
#     src_U_cube[src_x, src_y, :] = src_pol_spectrum.imag

#     src_Q_cube = gaussian_filter(
#         src_Q_cube, (beam_size_pix / 2.35, beam_size_pix / 2.35, 0), mode="wrap"
#     )
#     src_U_cube = gaussian_filter(
#         src_U_cube, (beam_size_pix / 2.35, beam_size_pix / 2.35, 0), mode="wrap"
#     )
#     scale_factor = (
#         np.max(np.sqrt(src_Q_cube**2 + src_U_cube**2)) / src_flux
#     )  # Renormalizing flux after convolution
#     src_Q_cube = src_Q_cube / scale_factor
#     src_U_cube = src_U_cube / scale_factor

#     diffuse_Q_cube = np.tile(
#         diffuse_pol_spectrum.real[np.newaxis, np.newaxis, :], (N_side, N_side, 1)
#     )
#     diffuse_U_cube = np.tile(
#         diffuse_pol_spectrum.imag[np.newaxis, np.newaxis, :], (N_side, N_side, 1)
#     )

#     rng = np.random.default_rng(20200422)
#     noise_Q_cube = rng.normal(scale=noise_amplitude, size=src_Q_cube.shape)
#     noise_U_cube = rng.normal(scale=noise_amplitude, size=src_Q_cube.shape)
#     noise_Q_cube = gaussian_filter(
#         noise_Q_cube, (beam_size_pix / 2.35, beam_size_pix / 2.35, 0), mode="wrap"
#     )
#     noise_U_cube = gaussian_filter(
#         noise_U_cube, (beam_size_pix / 2.35, beam_size_pix / 2.35, 0), mode="wrap"
#     )
#     scale_factor = (
#         np.std(noise_Q_cube) / noise_amplitude
#     )  # Renormalizing flux after convolution
#     noise_Q_cube = noise_Q_cube / scale_factor
#     noise_U_cube = noise_U_cube / scale_factor

#     Q_cube = src_Q_cube + noise_Q_cube + diffuse_Q_cube
#     U_cube = src_U_cube + noise_U_cube + diffuse_U_cube

#     header = pf.Header()
#     header["BITPIX"] = -32
#     header["NAXIS"] = 3
#     header["NAXIS1"] = N_side
#     header["NAXIS2"] = N_side
#     header["NAXIS3"] = freq_arr.size
#     header["CTYPE1"] = "RA---SIN"
#     header["CRVAL1"] = 90
#     header["CDELT1"] = -1.0 / 3600.0
#     header["CRPIX1"] = 1
#     header["CUNIT1"] = "deg"

#     header["CTYPE2"] = "DEC--SIN"
#     header["CRVAL2"] = 0
#     header["CDELT2"] = 1.0 / 3600.0
#     header["CRPIX2"] = 1
#     header["CUNIT2"] = "deg"

#     header["CTYPE3"] = "FREQ"
#     header["CRVAL3"] = freq_arr[0]
#     header["CDELT3"] = freq_arr[1] - freq_arr[0]
#     header["CRPIX3"] = 1
#     header["CUNIT3"] = "Hz"

#     header["BUNIT"] = "Jy/beam"

#     if not THREED_PATH.exists():
#         THREED_PATH.mkdir(parents=True)

#     pf.writeto(
#         THREED_PATH / "Q_cube.fits", np.transpose(Q_cube), header=header, overwrite=True
#     )
#     pf.writeto(
#         THREED_PATH / "U_cube.fits", np.transpose(U_cube), header=header, overwrite=True
#     )
#     with open(THREED_PATH / "freqHz.txt", "w") as f:
#         for freq in freq_arr:
#             f.write("{:}\n".format(freq))

#     with open(THREED_PATH / "sim_truth.txt", "w") as f:
#         f.write("Point source:\n")
#         f.write("RM = {} rad/m^2\n".format(src_RM))
#         f.write("Intrsinsic polarization angle = {} deg\n".format(src_pol_angle_deg))
#         f.write("Polarized Flux = {} Jy/beam\n".format(src_flux))
#         f.write("x position = {} pix\n".format(src_x))
#         f.write("y position = {} pix\n".format(src_y))
#         f.write("\n")
#         f.write("Diffuse emission:\n")
#         f.write("RM = {} rad/m^2\n".format(diffuse_RM))
#         f.write(
#             "Intrsinsic polarization angle = {} deg\n".format(diffuse_pol_angle_deg)
#         )
#         f.write("Polarized Flux = {} Jy/beam\n".format(diffuse_flux))
#         f.write("\n")
#         f.write("Other:\n")
#         f.write("Actual error per channel = {} Jy/beam\n".format(noise_amplitude))
#         f.write("Beam FWHM = {} pix\n".format(beam_size_pix))


# class test_RMtools(unittest.TestCase):
#     def setUp(self):
#         # Clean up old simulations to prevent interference with new runs.
#         N_chan = 288
#         self.freq_arr = np.linspace(800e6, 1088e6, num=N_chan)
#         self.models = (1, 2, 3, 4, 5, 7, 11)
#         self.sampler = "nestle"

#     def test_a1_1D_synth_runs(self):
#         create_1D_data(self.freq_arr)
#         returncode = subprocess.call(
#             f"rmsynth1d '{(ONED_PATH/'simsource.dat').as_posix()}' -l 600 -d 3 -S -i",
#             shell=True,
#         )
#         self.assertEqual(returncode, 0, "RMsynth1D failed to run.")

#     def test_a2_1D_synth_values(self):
#         mDict = json.load(open(ONED_PATH / "simsource_RMsynth.json", "r"))
#         refDict = json.load(open(TEST_PATH / "RMsynth1D_referencevalues.json", "r"))
#         for key in mDict.keys():
#             if (key == "polyCoefferr") or key == "polyCoeffs":
#                 ref_values = refDict[key].split(",")
#                 test_values = mDict[key].split(",")
#                 for ref, test in zip(ref_values, test_values):
#                     self.assertAlmostEqual(
#                         float(test),
#                         float(ref),
#                         places=3,
#                         msg=f"Key {key} differs from expectation",
#                     )
#             elif type(mDict[key]) == str or refDict[key] == 0:
#                 self.assertEqual(
#                     mDict[key], refDict[key], "{} differs from expectation.".format(key)
#                 )
#             else:
#                 self.assertTrue(
#                     np.abs((mDict[key] - refDict[key]) / refDict[key]) < 1e-3,
#                     "{} differs from expectation.".format(key),
#                 )

#     def test_c_3D_synth(self):
#         create_3D_data(self.freq_arr)
#         returncode = subprocess.call(
#             f"rmsynth3d '{(THREED_PATH/'Q_cube.fits').as_posix()}' '{(THREED_PATH/'U_cube.fits').as_posix()}' '{(THREED_PATH/'freqHz.txt').as_posix()}' -l 300 -d 10",
#             shell=True,
#         )
#         self.assertEqual(returncode, 0, "RMsynth3D failed to run.")
#         header = pf.getheader(THREED_PATH / "FDF_tot_dirty.fits")
#         self.assertEqual(header["NAXIS"], 3, "Wrong number of axes in output?")
#         self.assertEqual(
#             (header["NAXIS1"], header["NAXIS2"]),
#             (100, 100),
#             "Image plane has wrong dimensions!",
#         )
#         self.assertEqual(
#             header["NAXIS3"], 61, "Number of output FD planes has changed."
#         )

#     def test_b1_1D_clean(self):
#         if not (ONED_PATH / "simsource_RMsynth.dat").exists():
#             self.skipTest("Could not test 1D clean; 1D synth failed first.")
#         returncode = subprocess.call(
#             f"rmclean1d '{(ONED_PATH/'simsource.dat').as_posix()}' -n 11 -S", shell=True
#         )
#         self.assertEqual(returncode, 0, "RMclean1D failed to run.")

#     def test_b2_1D_clean_values(self):
#         mDict = json.load(open(ONED_PATH / "simsource_RMclean.json", "r"))
#         refDict = json.load(open(TEST_PATH / "RMclean1D_referencevalues.json", "r"))
#         for key in mDict.keys():
#             self.assertTrue(
#                 np.abs((mDict[key] - refDict[key]) / refDict[key]) < 1e-3,
#                 "{} differs from expectation.".format(key),
#             )

#     def test_d_3D_clean(self):
#         if not (THREED_PATH / "FDF_tot_dirty.fits").exists():
#             self.skipTest("Could not test 3D clean; 3D synth failed first.")
#         returncode = subprocess.call(
#             f"rmclean3d '{(THREED_PATH/'FDF_tot_dirty.fits').as_posix()}' '{(THREED_PATH/'RMSF_tot.fits').as_posix()}' -n 10",
#             shell=True,
#         )
#         self.assertEqual(returncode, 0, "RMclean3D failed to run.")
#         # what else?

#     def test_e_1Dsynth_fromFITS(self):
#         if not (THREED_PATH / "Q_cube.fits").exists():
#             create_3D_data(self.freq_arr)
#         returncode = subprocess.call(
#             f"rmsynth1dFITS '{(THREED_PATH/'Q_cube.fits').as_posix()}' '{(THREED_PATH/'U_cube.fits').as_posix()}'  25 25 -l 600 -d 3 -S",
#             shell=True,
#         )
#         self.assertEqual(returncode, 0, "RMsynth1D_fromFITS failed to run.")

#     def test_f1_QUfitting(self):
#         if not (ONED_PATH / "simsource.dat").exists():
#             create_1D_data(self.freq_arr)

#         local_models = Path("models_ns")
#         if not local_models.exists():
#             shutil.copytree(TEST_PATH / ".." / "RMtools_1D" / "models_ns", local_models)

#         for model in self.models:
#             returncode = subprocess.call(
#                 f"qufit simdata/1D/simsource.dat --sampler {self.sampler} -m {model}",
#                 shell=True,
#             )

#         self.assertEqual(returncode, 0, "QU fitting failed to run.")
#         shutil.rmtree(local_models)

#     def _test_f2_QUfit_values(self):
#         # I have temporarily disabled this test because it causes a lot of problems
#         # with values not being consistent across different runs.
#         err_limit = 0.05  # 5%

#         for model in self.models:
#             mDict = json.load(
#                 open(ONED_PATH / f"simsource_m{model}_{self.sampler}.json", "r")
#             )

#             refDict = json.load(
#                 open(
#                     TEST_PATH
#                     / f"QUfit_referencevalues/simsource_m{model}_{self.sampler}.json",
#                     "r",
#                 )
#             )

#             # The QU-fitting code has internal randomness that I can't control. So every run
#             # will produce slightly different results. I want to assert that these differences
#             # are below some limit.
#             for key, val in refDict.items():
#                 if isinstance(val, str):
#                     continue
#                 if isinstance(val, list):
#                     for i, v in enumerate(val):
#                         if isinstance(v, str):
#                             continue
#                         self.assertTrue(
#                             abs(mDict[key][i] - v) / abs(v) < err_limit,
#                             f"values[{i}] of {key} of model {model} differs from expectation.",
#                         )
#                 else:
#                     self.assertTrue(
#                         abs(mDict[key] - val) / abs(val) < err_limit,
#                         f"{key} of model {model} differs from expectation.",
#                     )


# if __name__ == "__main__":
#     os.chdir(os.path.dirname(os.path.realpath(__file__)))
#     if (TEST_PATH / "simdata").exists():
#         shutil.rmtree("simdata")

#     print("\nUnit tests running.")
#     print("Test data inputs and outputs can be found in {}\n\n".format(os.getcwd()))

#     unittest.TestLoader.sortTestMethodsUsing = None
#     suite = unittest.TestLoader().loadTestsFromTestCase(test_RMtools)
#     unittest.TextTestRunner(verbosity=2).run(suite)



# # =============================================================================#
# #                          MODEL DEFINITION FILE                              #
# # =============================================================================#
# import bilby
# import numpy as np


# # -----------------------------------------------------------------------------#
# # Function defining the model.                                                #
# #                                                                             #
# #  pDict       = Dictionary of parameters, created by parsing inParms, below. #
# #  lamSqArr_m2 = Array of lambda-squared values                               #
# #  quArr       = Complex array containing the Re and Im spectra.              #
# # -----------------------------------------------------------------------------#
# def model(pDict, lamSqArr_m2):
#     """

#     Single Faraday component with differential Faraday rotation
#     "Burn slab"

#     Ref:
#     Burn (1966) Eq 18; with N >> (H_r/2H_z^0)^2
#     Sokoloff et al. (1998) Eq 3
#     O'Sullivan et al. (2012) Eq 9
#     Ma et al. (2019a) Eq 12

#     """

#     # Calculate the complex fractional q and u spectra
#     # fmt: off
#     pArr = pDict["fracPol"] * np.ones_like(lamSqArr_m2)
#     quArr = (pArr * np.exp( 2j * (np.radians(pDict["psi0_deg"]) +
#                                   (0.5*pDict["deltaRM_radm2"] +
#                                    pDict["RM_radm2"]) * lamSqArr_m2))
#              * np.sin(pDict["deltaRM_radm2"] * lamSqArr_m2) /
#                (pDict["deltaRM_radm2"] * lamSqArr_m2))
#     # fmt: on

#     return quArr


# # -----------------------------------------------------------------------------#
# # Priors for the above model.                                                 #
# # See https://lscsoft.docs.ligo.org/bilby/prior.html for details.             #
# #                                                                             #
# # -----------------------------------------------------------------------------#
# priors = {
#     "fracPol": bilby.prior.Uniform(
#         minimum=0.0, maximum=1.0, name="fracPol", latex_label=r"$p$"
#     ),
#     "psi0_deg": bilby.prior.Uniform(
#         minimum=0,
#         maximum=180.0,
#         name="psi0_deg",
#         latex_label=r"$\psi_0$ (deg)",
#         boundary="periodic",
#     ),
#     "RM_radm2": bilby.prior.Uniform(
#         minimum=-1100.0,
#         maximum=1100.0,
#         name="RM_radm2",
#         latex_label=r"RM (rad m$^{-2}$)",
#     ),
#     "deltaRM_radm2": bilby.prior.Uniform(
#         minimum=0.0,
#         maximum=100.0,
#         name="deltaRM_radm2",
#         latex_label=r"$\Delta{RM}$ (rad m$^{-2}$)",
#     ),
# }



"""Logging module for arrakis"""

from __future__ import annotations

import io
import logging


class TqdmToLogger(io.StringIO):
    """
    Output stream for TQDM which will output to logger module instead of
    the StdOut.
    """

    logger = None
    level = None
    buf = ""

    def __init__(self, logger: logging.Logger, level: int = logging.INFO):
        super().__init__()
        self.logger = logger
        self.level = level or logging.INFO

    def write(self, buf):
        self.buf = buf.strip("\r\n\t ")

    def flush(self):
        if self.logger is not None and self.level is not None:
            self.logger.log(self.level, self.buf)


class CustomFormatter(logging.Formatter):
    format_str = "%(module)s.%(funcName)s: %(message)s"

    FORMATS = {  # noqa: RUF012
        logging.DEBUG: f"%(levelname)s {format_str}",
        logging.INFO: f"%(levelname)s {format_str}",
        logging.WARNING: f"%(levelname)s {format_str}",
        logging.ERROR: f"%(levelname)s {format_str}",
        logging.CRITICAL: f"%(levelname)s {format_str}",
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, "%Y-%m-%d %H:%M:%S")
        return formatter.format(record)


def get_logger(
    name: str = "rmtools-lite", attach_handler: bool = True
) -> logging.Logger:
    """Will construct a logger object.

    Args:
        name (str, optional): Name of the logger to attempt to use. This is ignored if in a prefect flowrun. Defaults to 'arrakis'.
        attach_handler (bool, optional): Attacjes a custom StreamHandler. Defaults to True.

    Returns:
        logging.Logger: The appropriate logger
    """
    logging.captureWarnings(True)
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)

    if attach_handler:
        # Create console handler and set level to debug
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)

        # Add formatter to ch
        ch.setFormatter(CustomFormatter())
        logger.addHandler(ch)

    return logger


logger = get_logger()
